import math
from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict, List, Callable, Optional, Tuple, Union

from qrogue.game import target_factory
from qrogue.game.logic import Message
from qrogue.game.logic.actors import Robot, Riddle
from qrogue.game.logic.collectibles import Instruction
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.game.world.tiles import Enemy as EnemyTile, Tile, Floor, Decoration, Teleport, FogOfWar, Void, Invalid, \
    Door, Wall, HallwayEntrance, Riddler, Challenger, Boss, Collectible, Message as MessageTile
from qrogue.util import CommonQuestions, MapConfig, Logger, CheatConfig, RandomManager


class AreaType(Enum):
    Invalid = -1
    Placeholder = 0
    Hallway = 1
    EmptyRoom = 2
    MetaRoom = 3

    SpawnRoom = 10
    WildRoom = 11       # contains mostly normal Enemies/Puzzles
    # ShopRoom = 12
    RiddleRoom = 13     # room is built around its Riddle(s)
    GateRoom = 14       # room contains mainly a Gate and other Tiles to support it
    BossRoom = 15       # room contains a Boss-Puzzle and other supporting Tiles
    TreasureRoom = 16   # room contains mainly some sort of Treasure/special Pickup and supporting Tiles
    ChallengeRoom = 17  # room is built around its Challenge(s)
    PauseRoom = 18      # room doesn't contain much interactive Tiles, mainly some Energy pickups
    StoryRoom = 19      # room is mainly there for story purposes, e.g. containing Messages
    PhantomRoom = 20    # a special hidden room that cannot be accessed via normal means (e.g. only via tunneling)

    @staticmethod
    def values(include_pseudo_areas: Optional[bool] = None, only_rooms: Optional[bool] = None) -> List["AreaType"]:
        if include_pseudo_areas is None:  # todo: method is not used
            include_pseudo_areas = False
        if only_rooms is None:
            only_rooms = True

        areas = [AreaType.SpawnRoom, AreaType.WildRoom, AreaType.RiddleRoom, AreaType.GateRoom,
                 AreaType.BossRoom, AreaType.TreasureRoom]
        if not only_rooms:
            areas += [AreaType.Hallway]
        if include_pseudo_areas:
            areas += [AreaType.Invalid, AreaType.Placeholder, AreaType.EmptyRoom, AreaType.MetaRoom]
        return areas


class Area(ABC):
    __ID = 1
    __FOG = FogOfWar()
    __VOID = Void()
    UNIT_WIDTH = MapConfig.room_width()  # todo change visibility of UNIT_WIDTH?
    UNIT_HEIGHT = MapConfig.room_height()
    MID_X = MapConfig.room_mid_x()
    MID_Y = MapConfig.room_mid_y()

    @staticmethod
    def void() -> Void:
        return Area.__VOID

    def __init__(self, type_: AreaType, tile_matrix: "list[list[Tile]]"):
        self.__id = Area.__ID
        Area.__ID += 1

        self.__type = type_
        self.__tiles = tile_matrix
        self.__width = len(tile_matrix[0])
        self.__height = len(tile_matrix)

        self.__is_in_sight = False
        self.__is_visible = False
        self.__was_visited = False

    @property
    def id(self) -> int:
        return self.__id

    @property
    def is_visible(self) -> bool:
        return self.__is_visible or CheatConfig.revealed_map()

    @property
    def _is_visible_value(self) -> bool:
        return self.__is_visible

    @property
    def is_in_sight(self) -> bool:
        return self.__is_in_sight

    @property
    def type(self) -> AreaType:
        return self.__type

    def _set_tile(self, tile: Tile, x: int, y: int) -> bool:
        """

        :param tile: the Tile we want to place
        :param x: horizontal position of the Tile
        :param y: vertical position of the Tile
        :return: whether we could place the Tile at the given position or not
        """
        if 0 <= x < Area.UNIT_WIDTH and 0 <= y < Area.UNIT_HEIGHT:
            self.__tiles[y][x] = tile
            return True
        return False

    def at(self, x: int, y: int, force: bool = False) -> Tile:
        """

        :param x: horizontal position of the Tile we want to know
        :param y: vertical position of the Tile we want to know
        :param force: whether we return the real Tile or not (e.g. invisible Rooms usually return Fog)
        :return: the Tile at the requested position
        """
        if 0 <= x < self.__width and 0 <= y < self.__height:
            if self.is_visible or force:
                return self.__tiles[y][x]
            else:
                if self.is_in_sight:
                    return Area.__FOG
                else:
                    return Area.void()
        else:
            return Invalid()

    def get_row_str(self, row: int) -> str:
        if row >= len(self.__tiles):
            return "".join([Invalid().get_img()] * Area.UNIT_WIDTH)

        if self.is_visible:
            tiles = [t.get_img() for t in self.__tiles[row]]
            return "".join(tiles)
        elif self.is_in_sight:
            fog_str = Area.__FOG.get_img()
            return fog_str * self.__width
        else:
            void_str = Area.__VOID.get_img()
            return void_str * self.__width

    def make_visible(self):
        self.__is_in_sight = True
        self.__is_visible = True

    def in_sight(self):
        self.__is_in_sight = True

    def enter(self, direction: Direction):
        self.__is_visible = True
        self.__was_visited = True

    def leave(self, direction: Direction):
        pass

    def __str__(self):
        return "?"


class Hallway(Area):
    @staticmethod
    def is_first(direction: Direction):
        """
        Determines whether a Hallway in the Direction direction from the perspective of a room is the first or second
        Room of the Hallway
        :param direction: Direction from Room to Hallway
        :return: True if Hallway is to the East or South of Room, false otherwise
        """
        return direction in [Direction.East, Direction.South]

    def __init__(self, door: Door):
        self.__door = door
        self.__hide = door.is_event_locked
        self.__room1: Optional[Room] = None
        self.__room2: Optional[Room] = None
        if self.is_horizontal():
            missing_half = int((Area.UNIT_WIDTH - 3) / 2)
            row: List[Tile] = [Void()] * missing_half + [Wall(), door, Wall()] + [Void()] * missing_half
            super(Hallway, self).__init__(AreaType.Hallway, [row])
        else:
            missing_half = int((Area.UNIT_HEIGHT - 3) / 2)
            tiles: List[List[Tile]] = [[Void()]] * missing_half + [[Wall()], [door], [Wall()]] + [
                [Void()]] * missing_half
            super(Hallway, self).__init__(AreaType.Hallway, tiles)

    @property
    def door(self) -> Door:
        return self.__door

    def set_room(self, room: "Room", direction: Direction):
        """

        :param room:
        :param direction:  in which Direction the Hallways is from room's point of view
        :return:
        """
        if Hallway.is_first(direction):
            self.__room1 = room
        else:
            self.__room2 = room

    def connects_horizontally(self) -> bool:
        """

        :return: whether the Hallway connects two Rooms horizontally or not
        """
        return self.__door.direction.is_horizontal()

    def is_horizontal(self) -> bool:
        """

        :return: True if the Hallway is a row, False if it is a column
        """
        return not self.connects_horizontally()

    def make_visible(self):
        super(Hallway, self).make_visible()
        r1id, r2id = None, None
        if self.__room1:
            r1id = self.__room1.id
            self.__room1.in_sight()
        if self.__room2:
            r2id = self.__room2.id
            self.__room2.in_sight()

        if r1id is None and r2id is None:
            Logger.instance().debug(f"#hw{self.id}: room1 and room2 is None!", from_pycui=False)
        elif r1id is None:
            Logger.instance().debug(f"#hw{self.id}: room2=#r{r2id} but room1 is None!", from_pycui=False)
        elif r2id is None:
            Logger.instance().debug(f"#hw{self.id}: room1=#r{r1id} but room2 is None!", from_pycui=False)

    def get_row_str(self, row: int) -> str:
        if self.__hide:
            if self.__door.check_event():
                if self.__room1.is_visible or self.__room2.is_visible:
                    self.make_visible()
                # elif self.__room1.in_sight or self.__room2.in_sight:
                #    self.in_sight()
                self.__hide = False
        return super(Hallway, self).get_row_str(row)

    def in_sight(self):
        if not self.__hide:
            self.make_visible()

    def enter(self, direction: Direction):
        self.__room1.make_visible()
        self.__room2.make_visible()

    def room(self, first: bool):
        if first:
            return self.__room1
        else:
            return self.__room2

    def __str__(self) -> str:
        if self.is_horizontal():
            orientation = "-"
        else:
            orientation = "|"
        return f"HW{self.id}: {self.__room1} {orientation} {self.__room2}"


class Room(Area):
    INNER_WIDTH = Area.UNIT_WIDTH - 2  # width inside the room, i.e. without walls and hallways
    INNER_HEIGHT = Area.UNIT_HEIGHT - 2  # height inside the room, i.e. without walls and hallways
    INNER_MID_X = int(INNER_WIDTH / 2)
    INNER_MID_Y = int(INNER_HEIGHT / 2)

    @staticmethod
    def _set(tile_list: List[Tile], x: int, y: int, tile: Tile) -> bool:
        index = y * Room.INNER_WIDTH + x
        if 0 <= index < len(tile_list):
            tile_list[index] = tile
            return True
        return False

    @staticmethod
    def mid(inner: bool = True) -> Coordinate:
        if inner:
            return Coordinate(Room.INNER_MID_X, Room.INNER_MID_Y)
        else:
            return Coordinate(Area.MID_X, Area.MID_Y)

    @staticmethod
    def coordinate_to_index(pos: Union[Coordinate, Tuple[int, int]], in_room: bool = True) -> int:
        """

        :param pos: either Coordinate or Tuple of x and y position
        :param in_room: whether pos is given as in-room position or not (i.e. skip ring of walls or not)
        :return: index of a row-based tile list corresponding to the given position
        """
        if in_room:
            width = Room.INNER_WIDTH
        else:
            width = Room.UNIT_WIDTH
        if isinstance(pos, Coordinate):
            x, y = pos.x, pos.y
        else:
            x, y = pos
        return x + y * width

    @staticmethod
    def index_to_coordinate(index: int, in_room: bool = True) -> Coordinate:
        """

        :param index: index of a row-based tile list
        :param in_room: whether index is given as in-room position or not (i.e. skip ring of walls or not)
        :return: position corresponding to the given index of a row-based tile list
        """
        if in_room:
            width, height = Room.INNER_WIDTH, Room.INNER_HEIGHT
        else:
            width, height = Room.UNIT_WIDTH, Room.UNIT_HEIGHT
        assert 0 <= index < width * height, f"Index = {index} out of range: [0, {width * height}["
        return Coordinate(index % width, int(index / width))

    @staticmethod
    def direction_to_hallway_entrance_pos(direction: Direction, in_room: bool = True) -> Coordinate:
        """

        :param direction: side of the room where the hallway entrance is
        :param in_room: whether the return value is an in-room position or not (i.e. skip ring of walls or not)
        :return:
        """
        if in_room:
            # we have to use adapted middles since there is no ring of walls yet
            in_room_mid_x = Room.MID_X - 1
            in_room_mid_y = Room.MID_Y - 1
        else:
            in_room_mid_x = Room.MID_X
            in_room_mid_y = Room.MID_Y
        # up and left destructively interfere with +1, resulting in 0
        # while right and bottom constructively interfere with +1, resulting in 2
        return Coordinate(in_room_mid_x * (1 + direction.x), in_room_mid_y * (1 + direction.y))

    @staticmethod
    def get_empty_room_tile_list() -> "list of Tiles":
        return [Floor()] * (Room.INNER_WIDTH * Room.INNER_HEIGHT)

    @staticmethod
    def dic_to_tile_list(tile_dic: Dict[Coordinate, Tile]) -> List[Tile]:
        tile_list = Room.get_empty_room_tile_list()
        if tile_dic:
            for coordinate, tile in tile_dic.items():
                index = Room.coordinate_to_index(coordinate)
                if 0 <= index < len(tile_list):
                    tile_list[index] = tile
                else:
                    Logger.instance().throw(IndexError(f"Invalid Index ({index} for tile_list of length "
                                                       f"{len(tile_list)}!"))
        return tile_list

    def __init__(self, type_: AreaType, tile_list: List[Tile], north_hallway: Optional[Hallway] = None,
                 east_hallway: Optional[Hallway] = None, south_hallway: Optional[Hallway] = None,
                 west_hallway: Optional[Hallway] = None):
        tiles: List[List[Tile]] = []
        room_top = [Wall()] * Area.UNIT_WIDTH
        tiles.append(room_top)

        # fence the room tiles with Walls
        for y in range(Room.INNER_HEIGHT):
            row = [Wall()]
            for x in range(Room.INNER_WIDTH):
                if tile_list is not None:
                    index = x + y * Room.INNER_WIDTH
                    if index < len(tile_list):
                        row.append(tile_list[index])
                    else:
                        row.append(Floor())
                else:
                    row.append(Floor())
            row.append(Wall())
            tiles.append(row)

        room_bottom = [Wall()] * Area.UNIT_WIDTH
        tiles.append(room_bottom)

        self.__hallways: Dict[Direction, Hallway] = {}
        if north_hallway is not None and not north_hallway.connects_horizontally():
            tiles[0][Area.MID_X] = HallwayEntrance(north_hallway.door)
            self.__hallways[Direction.North] = north_hallway
            north_hallway.set_room(self, Direction.North)

        if east_hallway is not None and east_hallway.connects_horizontally():
            tiles[Area.MID_Y][Area.UNIT_WIDTH - 1] = HallwayEntrance(east_hallway.door)
            self.__hallways[Direction.East] = east_hallway
            east_hallway.set_room(self, Direction.East)

        if south_hallway is not None and not south_hallway.connects_horizontally():
            tiles[Area.UNIT_HEIGHT - 1][Area.MID_X] = HallwayEntrance(south_hallway.door)
            self.__hallways[Direction.South] = south_hallway
            south_hallway.set_room(self, Direction.South)

        if west_hallway is not None and west_hallway.connects_horizontally():
            tiles[Area.MID_Y][0] = HallwayEntrance(west_hallway.door)
            self.__hallways[Direction.West] = west_hallway
            west_hallway.set_room(self, Direction.West)

        super(Room, self).__init__(type_, tiles)

    def _set_hallway(self, north_hallway: Hallway = None, east_hallway: Hallway = None, south_hallway: Hallway = None,
                     west_hallway: Hallway = None):
        if north_hallway is not None and not north_hallway.connects_horizontally():
            self._set_tile(Floor(), Area.MID_X, 0)
            self.__hallways[Direction.North] = north_hallway
            north_hallway.set_room(self, Direction.North)

        if east_hallway is not None and east_hallway.connects_horizontally():
            self._set_tile(Floor(), Area.UNIT_WIDTH - 1, Area.MID_Y)
            self.__hallways[Direction.East] = east_hallway
            east_hallway.set_room(self, Direction.East)

        if south_hallway is not None and not south_hallway.connects_horizontally():
            self._set_tile(Floor(), Area.MID_X, Area.UNIT_HEIGHT - 1)
            self.__hallways[Direction.South] = south_hallway
            south_hallway.set_room(self, Direction.South)

        if west_hallway is not None and west_hallway.connects_horizontally():
            self._set_tile(Floor(), 0, Area.MID_Y)
            self.__hallways[Direction.West] = west_hallway
            west_hallway.set_room(self, Direction.West)

    def get_hallway(self, direction: Direction, throw_error: bool = True) -> Optional[Hallway]:
        if direction in self.__hallways:
            return self.__hallways[direction]
        elif throw_error:
            dic = [(str(self.__hallways[k]) + "\n") for k in self.__hallways]
            Logger.instance().error(f"Invalid hallway access for {self}:\n{direction} not in {dic}", show=False,
                                    from_pycui=False)
        return None

    def make_visible(self):
        super(Room, self).make_visible()
        for key in self.__hallways:
            hallway = self.__hallways[key]
            hallway.in_sight()

    def at(self, x: int, y: int, force: bool = False, inside_only: bool = False) -> Tile:
        """

        :param x: horizontal position of the Tile we want to know
        :param y: vertical position of the Tile we want to know
        :param force: whether we return the real Tile or not (e.g. invisible Rooms usually return Fog)
        :param inside_only: whether we consider the whole room or only its content inside the static walls; interprets
                            x and y accordingly
        :return: the Tile at the requested position
        """
        if inside_only:
            if 0 <= x < Room.INNER_WIDTH and 0 <= y < Room.INNER_HEIGHT:
                # offset position by 1
                return super(Room, self).at(x + 1, y + 1, force)
            return Invalid()
        else:
            return super(Room, self).at(x, y, force)

    @abstractmethod
    def abbreviation(self) -> str:
        pass

    def to_string(self) -> str:
        text = str(self) + "\n"
        for y in range(Area.UNIT_HEIGHT):
            for x in range(Area.UNIT_WIDTH):
                tile = self.at(x, y, force=True)
                text += tile.get_img()
            text += "\n"
        return text

    def __str__(self) -> str:
        return self.abbreviation() + str(self.id)


class CopyAbleRoom(Room, ABC):
    @abstractmethod
    def copy(self, hw_dic: Dict[Direction, Hallway]) -> "CopyAbleRoom":
        pass


class MetaRoom(CopyAbleRoom):
    def __init__(self, load_map_callback: Callable[[str], None], orientation: Direction, message: Message,
                 level_to_load: str, mtype: str, num: int, blocking: bool = True, north_hallway: Hallway = None,
                 east_hallway: Hallway = None, south_hallway: Hallway = None, west_hallway: Hallway = None,
                 is_spawn: bool = False):
        self.__load_map_callback = load_map_callback
        self.__orientation = orientation
        self.__message = message
        self.__level_to_load = level_to_load
        self.__mtype = mtype
        self.__num = num
        self.__blocking = blocking
        self.__is_spawn = is_spawn

        if is_spawn and False:  # todo not used/wanted currently, maybe add back later
            super(MetaRoom, self).__init__(AreaType.MetaRoom, Room.get_empty_room_tile_list(), north_hallway,
                                           east_hallway, south_hallway, west_hallway)
        else:
            if blocking:
                if orientation.is_horizontal():
                    dash_decoration = Decoration('|')
                else:
                    dash_decoration = Decoration('-')
            else:
                dash_decoration = Floor()  # ignore it for non-blocking since it might be placed in front of a hallway

            type_decoration = Decoration(mtype)
            num_decoration = Decoration(str(num)[0])  # todo handle double digits

            tile_list = Room.get_empty_room_tile_list()
            msg_tile = MessageTile(self.__message, -1)
            level_tile = Teleport(self.__load_map_callback, level_to_load, None)

            if orientation is Direction.North:
                self._set(tile_list, 0, 0, type_decoration)
                self._set(tile_list, Room.INNER_MID_X, 0, dash_decoration)
                self._set(tile_list, Room.INNER_WIDTH - 1, 0, num_decoration)
                self._set(tile_list, 0, Room.INNER_HEIGHT - 1, msg_tile)
                self._set(tile_list, Room.INNER_WIDTH - 1, Room.INNER_HEIGHT - 1, level_tile)
                coordinate = Coordinate(0, Room.INNER_MID_Y - 1)
                addition = Direction.East

            elif orientation is Direction.East:
                self._set(tile_list, Room.INNER_WIDTH - 1, 0, type_decoration)
                self._set(tile_list, Room.INNER_WIDTH - 1, Room.INNER_MID_Y, dash_decoration)
                self._set(tile_list, Room.INNER_WIDTH - 1, Room.INNER_HEIGHT - 1, num_decoration)
                self._set(tile_list, 0, 0, msg_tile)
                self._set(tile_list, 0, Room.INNER_HEIGHT - 1, level_tile)
                coordinate = Coordinate(Room.INNER_MID_X + 1, 0)
                addition = Direction.South

            elif orientation is Direction.South:
                self._set(tile_list, 0, Room.INNER_HEIGHT - 1, type_decoration)
                self._set(tile_list, Room.INNER_MID_X, Room.INNER_HEIGHT - 1, dash_decoration)
                self._set(tile_list, Room.INNER_WIDTH - 1, Room.INNER_HEIGHT - 1, num_decoration)
                self._set(tile_list, 0, 0, msg_tile)
                self._set(tile_list, Room.INNER_WIDTH - 1, 0, level_tile)
                coordinate = Coordinate(0, Room.INNER_MID_Y + 1)
                addition = Direction.East

            elif orientation is Direction.West:
                self._set(tile_list, 0, 0, type_decoration)
                self._set(tile_list, 0, Room.INNER_MID_Y, dash_decoration)
                self._set(tile_list, 0, Room.INNER_HEIGHT - 1, num_decoration)
                self._set(tile_list, Room.INNER_WIDTH - 1, 0, msg_tile)
                self._set(tile_list, Room.INNER_WIDTH - 1, Room.INNER_HEIGHT - 1, level_tile)
                coordinate = Coordinate(Room.INNER_MID_X - 1, 0)
                addition = Direction.South

            else:
                raise SyntaxError(f"Invalid orientation for Hallway: {orientation}.")

            if blocking:
                counter = 0
                while counter < 5 and self._set(tile_list, coordinate.x, coordinate.y, Wall()):
                    coordinate += addition
                    counter += 1
            super(MetaRoom, self).__init__(AreaType.MetaRoom, tile_list, north_hallway, east_hallway, south_hallway,
                                           west_hallway)

    def abbreviation(self) -> str:
        return "MR"

    def copy(self, hw_dic: Dict[Direction, Hallway]) -> CopyAbleRoom:
        # if we wouldn't copy it we could very easily get errors because the hallways
        # will be set for all the layout positions that reference this room and hence
        # can ultimately lead to non-existing rooms
        if hw_dic:
            new_room = MetaRoom(self.__load_map_callback, self.__orientation, self.__message, self.__level_to_load,
                                self.__mtype, self.__num, self.__blocking, hw_dic[Direction.North],
                                hw_dic[Direction.East], hw_dic[Direction.South], hw_dic[Direction.West],
                                self.__is_spawn)
        else:
            raise SyntaxError("Room without hallway!")

        if self._is_visible_value:
            # don't use Room's implementation but Area's
            super(CopyAbleRoom, new_room).make_visible()
        elif self.is_in_sight:
            new_room.in_sight()
        return new_room


class CustomRoom(CopyAbleRoom):
    def __init__(self, type_: AreaType, tile_matrix: Optional[List[List[Tile]]],
                 north_hallway: Optional[Hallway] = None, east_hallway: Optional[Hallway] = None,
                 south_hallway: Optional[Hallway] = None, west_hallway: Optional[Hallway] = None):
        self.__tile_matrix = tile_matrix
        tile_list = []
        for row in tile_matrix:
            tile_list += row
        super(CustomRoom, self).__init__(type_, tile_list, north_hallway, east_hallway, south_hallway, west_hallway)

    def abbreviation(self) -> str:
        return "CR"

    def copy(self, hw_dic: Dict[Direction, Hallway]) -> CopyAbleRoom:
        # if we wouldn't copy it we could very easily get errors because the hallways
        # will be set for all the layout positions that reference this room and hence
        # can ultimately lead to non-existing rooms
        if hw_dic:
            new_room = CustomRoom(self.type, self.__tile_matrix, hw_dic[Direction.North], hw_dic[Direction.East],
                                  hw_dic[Direction.South], hw_dic[Direction.West])
        else:
            # todo not sure if rooms without hallways should be legal
            new_room = CustomRoom(self.type, self.__tile_matrix)

        if self._is_visible_value:
            # don't use Room's implementation but Area's
            super(Room, new_room).make_visible()
        elif self.is_in_sight:
            new_room.in_sight()
        return new_room


class EmptyRoom(CopyAbleRoom):
    def __init__(self, hw_dic: Dict[Direction, Hallway]):
        north_hallway = None
        if Direction.North in hw_dic:
            north_hallway = hw_dic[Direction.North]

        east_hallway = None
        if Direction.East in hw_dic:
            east_hallway = hw_dic[Direction.East]

        south_hallway = None
        if Direction.South in hw_dic:
            south_hallway = hw_dic[Direction.South]

        west_hallway = None
        if Direction.West in hw_dic:
            west_hallway = hw_dic[Direction.West]

        super(EmptyRoom, self).__init__(AreaType.EmptyRoom, Room.get_empty_room_tile_list(), north_hallway,
                                        east_hallway, south_hallway, west_hallway)

    def abbreviation(self) -> str:
        return "ER"

    def copy(self, hw_dic: Dict[Direction, Hallway]) -> "CopyAbleRoom":
        return EmptyRoom(hw_dic)


class SpecialRoom(Room, ABC):
    # todo check if it should still apply that SpecialRooms can only have one Hallway - adapt Astar in LayoutGenerator
    # todo if that changes!
    # todo answering thoughts after some time: If they can have multiple entrances/exits Key placement gets much harder
    # todo we could not simply place keys in non-special rooms but would also have to check if 1st key is reachable from
    # todo SpawnRoom, 2nd key from SpawnRoom and one special room, ...
    def __init__(self, type_: AreaType, hallway: Hallway, direction: Direction, special_tile: Tile,
                 tile_dic: Optional[Dict[Coordinate, Tile]] = None):
        """

        :param hallway:
        :param direction: the Direction to the Hallway based on the Room's perspective
        :param special_tile: the tile this room is dedicated to, placed in the middle
        :param tile_dic:
        """
        if tile_dic is None: tile_dic = {}
        tile_dic[Room.mid()] = special_tile
        tile_list = Room.dic_to_tile_list(tile_dic)
        if Hallway.is_first(direction):
            if hallway.connects_horizontally():
                super(SpecialRoom, self).__init__(type_, tile_list, east_hallway=hallway)
            else:
                super(SpecialRoom, self).__init__(type_, tile_list, south_hallway=hallway)
        else:
            if hallway.connects_horizontally():
                super(SpecialRoom, self).__init__(type_, tile_list, west_hallway=hallway)
            else:
                super(SpecialRoom, self).__init__(type_, tile_list, north_hallway=hallway)


class SpawnRoom(CopyAbleRoom):
    def __init__(self, leave_map_callback: Callable[[], None],
                 tile_dic: Dict[Coordinate, Tile] = None, north_hallway: Hallway = None, east_hallway: Hallway = None,
                 south_hallway: Hallway = None, west_hallway: Hallway = None, place_teleporter: bool = True):
        room_mid = Coordinate(Room.INNER_MID_X, Room.INNER_MID_Y)
        if tile_dic is None:
            tile_dic = {}
        else:
            if room_mid in tile_dic and place_teleporter:
                Logger.instance().error("Specified tile_dic with non-empty room-center for SpawnRoom. Overriding it "
                                        "with Teleporter.", show=False, from_pycui=False)
        if place_teleporter:
            tile_dic[room_mid] = Teleport(self.__teleport_callback, MapConfig.back_map_string(), None)
        self.__tile_dic = tile_dic
        super(SpawnRoom, self).__init__(AreaType.SpawnRoom, Room.dic_to_tile_list(tile_dic), north_hallway,
                                        east_hallway, south_hallway, west_hallway)
        self.__leave_map = leave_map_callback
        self.__is_done = None

    def abbreviation(self):
        return "SR"

    def set_is_done_callback(self, is_done_callback: Callable[[], bool]):
        self.__is_done = is_done_callback

    def __teleport_callback(self, level_to_load: str, room: Optional[Coordinate]):
        if self.__is_done:
            if self.__is_done():
                self.__conditional_going_back(0)
            else:
                CommonQuestions.GoingBack.ask(self.__conditional_going_back)
        else:
            Logger.instance().error("is_done_callback not set yet!", show=False, from_pycui=False)

    def __conditional_going_back(self, confirmed: int):
        if confirmed == 0:
            self.__leave_map()

    def copy(self, hw_dic: Dict[Direction, Hallway]) -> "CopyAbleRoom":
        # don't place a teleporter because it's already in tile_dic if it should be placed
        return SpawnRoom(self.__leave_map, self.__tile_dic, hw_dic[Direction.North], hw_dic[Direction.East],
                         hw_dic[Direction.South], hw_dic[Direction.West], place_teleporter=False)


class BaseWildRoom(Room):
    def __init__(self, tile_list: List[Tile], get_tiles_by_id: Callable[[int], List[Tile]],
                 north_hallway: Optional[Hallway] = None, east_hallway: Optional[Hallway] = None,
                 south_hallway: Optional[Hallway] = None, west_hallway: Optional[Hallway] = None):
        self.get_tiles_by_id = get_tiles_by_id
        super(BaseWildRoom, self).__init__(AreaType.WildRoom, tile_list, north_hallway, east_hallway, south_hallway,
                                           west_hallway)

    def abbreviation(self):
        return "WR"


class WildRoom(BaseWildRoom):
    __NUM_OF_ENEMY_GROUPS = 4

    def __init__(self, factory: target_factory.EnemyFactory, chance: float = 0.6, north_hallway: Hallway = None,
                 east_hallway: Hallway = None, south_hallway: Hallway = None, west_hallway: Hallway = None):
        from qrogue.util import Config
        Config.check_reachability("WildRoom()")
        self.__dictionary = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [], 8: [], 9: []}
        rm = RandomManager.create_new(7)

        available_positions = []
        for y in range(Room.INNER_HEIGHT):
            available_positions += [Coordinate(x, y) for x in range(Room.INNER_WIDTH)]
        num_of_enemies = len(available_positions) * chance
        if rm.get_bool(msg="WR_enemyPlacementRounding"):
            num_of_enemies = math.floor(num_of_enemies)
        else:
            num_of_enemies = math.ceil(num_of_enemies)

        # here we could add other stuff (pickups, buttons?, ...) to the room and adapt available_positions accordingly

        tile_list = Room.get_empty_room_tile_list()
        for i in range(num_of_enemies):
            eid = rm.get_int(min_=0, max_=WildRoom.__NUM_OF_ENEMY_GROUPS + 1, msg="WildRoom_eid")
            enemy = EnemyTile(factory, self.__get_tiles_by_id, self.__update_entangled_tiles, eid)
            if eid > 0:
                self.__dictionary[eid].append(enemy)
            pos = rm.get_element(available_positions, remove=True, msg="WildRoom_epos")
            tile_list[Room.coordinate_to_index(pos)] = enemy

        super().__init__(tile_list, self.__get_tiles_by_id, north_hallway, east_hallway, south_hallway, west_hallway)

    def __get_tiles_by_id(self, eid: int) -> List[EnemyTile]:
        return self.__dictionary[eid]

    def __update_entangled_tiles(self, enemy: EnemyTile):
        self.__dictionary[enemy.eid].append(enemy)


class DefinedWildRoom(BaseWildRoom):
    def __init__(self, tile_list: List[Tile], north_hallway: Hallway = None, east_hallway: Hallway = None,
                 south_hallway: Hallway = None, west_hallway: Hallway = None):
        self.__dictionary = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [], 8: [], 9: []}
        for tile in tile_list:
            # 0s always collapse to "fight" so they don't need to be in the dictionary for entanglement
            if isinstance(tile, EnemyTile) and tile.eid > 0:
                self.__dictionary[tile.eid].append(tile)
        super().__init__(tile_list, self.__get_tiles_by_id, north_hallway=north_hallway, east_hallway=east_hallway,
                         south_hallway=south_hallway, west_hallway=west_hallway)

    def __get_tiles_by_id(self, id_: int) -> List[EnemyTile]:
        return self.__dictionary[id_]


class TreasureRoom(SpecialRoom):
    def __init__(self, treasure: Union[Collectible, Instruction], hallway: Hallway, direction: Direction,
                 tile_dic: Dict[Coordinate, Tile] = None):
        if isinstance(treasure, Instruction):
            super().__init__(AreaType.GateRoom, hallway, direction, Collectible(treasure, force_add=True), tile_dic)
        else:
            super().__init__(AreaType.TreasureRoom, hallway, direction, treasure, tile_dic)

    def abbreviation(self) -> str:
        return "TR"


class ChallengeRoom(SpecialRoom):
    def __init__(self, hallway: Hallway, direction: Direction, challenger: Challenger,
                 tile_dic: Dict[Coordinate, Tile] = None):
        super().__init__(AreaType.ChallengeRoom, hallway, direction, challenger, tile_dic)

    def abbreviation(self) -> str:
        return "TR"


class RiddleRoom(SpecialRoom):
    def __init__(self, hallway: Hallway, direction: Direction, riddle: Riddle,
                 open_riddle_callback: Callable[[Robot, Riddle], None], tile_dic: Dict[Coordinate, Tile] = None):
        super().__init__(AreaType.RiddleRoom, hallway, direction, Riddler(open_riddle_callback, riddle), tile_dic)

    def abbreviation(self):
        return "RR"


class BossRoom(SpecialRoom):
    def __init__(self, hallway: Hallway, direction: Direction, boss: Boss, tile_dic: Dict[Coordinate, Tile] = None):
        super().__init__(AreaType.BossRoom, hallway, direction, boss, tile_dic)

    def abbreviation(self) -> str:
        return "BR"


class Placeholder:
    class _AreaPlaceholder(Area):
        def __init__(self, code: int):
            if code == 0:
                # horizontal Hallway
                self.__has_full_row = True
            elif code == 1:
                # vertical Hallway
                self.__has_full_row = False
            else:
                # room
                self.__has_full_row = True
            super().__init__(AreaType.Placeholder, [[Area.void()]])

        def at(self, x: int, y: int, force: bool = False) -> Tile:
            return Area.void()

        def get_row_str(self, row: int) -> str:
            void_str = Area.void().get_img()
            if self.__has_full_row:
                return void_str * Area.UNIT_WIDTH
            else:
                return void_str

        def in_sight(self):
            pass

        def enter(self, direction: Direction):
            pass

    class _RoomPlaceHolder(Room):
        def __init__(self):
            super(Placeholder._RoomPlaceHolder, self).__init__(AreaType.Invalid, [])

        def abbreviation(self) -> str:
            return "PH"

        def at(self, x: int, y: int, force: bool = False, inside_only: bool = False) -> Tile:
            return Invalid()

    __HORIZONTAL = _AreaPlaceholder(0)
    __VERTICAL = _AreaPlaceholder(1)
    __ROOM = _AreaPlaceholder(2)

    @staticmethod
    def horizontal() -> _AreaPlaceholder:
        return Placeholder.__HORIZONTAL

    @staticmethod
    def vertical() -> _AreaPlaceholder:
        return Placeholder.__VERTICAL

    @staticmethod
    def pseudo_room() -> _AreaPlaceholder:
        return Placeholder.__ROOM

    @staticmethod
    def empty_room(hw_dic: Dict[Direction, "Hallway"]) -> "CopyAbleRoom":
        return EmptyRoom(hw_dic)

    @staticmethod
    def room() -> _RoomPlaceHolder:
        return Placeholder._RoomPlaceHolder()



import random
import sys
from typing import Tuple, List, Optional

from qrogue.game.world.dungeon_generator import QrogueLevelGenerator
from qrogue.game.world.dungeon_generator.wave_function_collapse import WFCManager
from qrogue.game.world.map import CallbackPack
from qrogue.management import QrogueCUI
from qrogue.util import Logger, Config, PyCuiConfig, PathConfig, GameplayConfig, Options


def __init_singletons(seed: int):
    Logger()
    Logger.instance().info(Config.get_log_head(seed), from_pycui=False)


def __parse_argument(argument: List[str], has_value: bool = False) -> Tuple[bool, Optional[str]]:
    for arg in argument:
        if arg in sys.argv:
            if has_value:
                i = sys.argv.index(arg)
                if i + 1 < len(sys.argv):
                    return True, sys.argv[i + 1]
            else:
                return True, None
    return False, None


def start_qrogue() -> int:
    Logger.print_to_console("Loading...")  # notify player that the game is loading

    __CONSOLE_ARGUMENT = ["--from-console", "-fc"]
    __DEBUG_ARGUMENT = ["--debug", "-d"]
    __SEED_ARGUMENT = ["--seed", "-s"]  # seed argument
    __TEST_LEVEL_ARGUMENT = ["--test-level", "-tl"]
    __GAME_DATA_PATH_ARGUMENT = ["--game-data", "-gd"]  # path argument
    __USER_DATA_PATH_ARGUMENT = ["--user-data", "-ud"]  # path argument
    # __CONTROLS_ARGUMENT = ["--controls", "-c"]          # int argument
    __SIMULATION_FILE_ARGUMENT = ["--simulation-path", "-sp"]  # path argument
    __VALIDATE_MAP_ARGUMENT = ["--validate-map", "-vm"]  # path argument
    __PLAY_LEVEL_ARGUMENT = ["--play-level", "-pl"]  # str argument
    __UPDATE_WFC = ["--update-wfc", "-wfc"]

    from_console, _ = __parse_argument(__CONSOLE_ARGUMENT)
    debugging, _ = __parse_argument(__DEBUG_ARGUMENT)
    test_level, _ = __parse_argument(__TEST_LEVEL_ARGUMENT)
    _, data_folder = __parse_argument(__GAME_DATA_PATH_ARGUMENT, has_value=True)
    _, user_data_folder = __parse_argument(__USER_DATA_PATH_ARGUMENT, has_value=True)
    has_simulation_path, simulation_path = __parse_argument(__SIMULATION_FILE_ARGUMENT, has_value=True)
    has_map_path, map_path = __parse_argument(__VALIDATE_MAP_ARGUMENT, has_value=True)
    _, seed = __parse_argument(__SEED_ARGUMENT, has_value=True)
    has_play_level, level2play = __parse_argument(__PLAY_LEVEL_ARGUMENT, has_value=True)
    has_update_wfc, _ = __parse_argument(__UPDATE_WFC)

    if has_update_wfc:
        update_wfc()
    elif has_map_path:
        if validate_map(map_path):
            return 0
        else:
            return 1
    else:
        if has_simulation_path:
            return simulate_game(simulation_path, from_console, debugging, data_folder, user_data_folder)
        elif has_play_level:
            return play_level(level2play, debugging, data_folder, seed)
        else:
            return start_game(from_console, debugging, test_level, data_folder, user_data_folder, seed)


def setup_game(game_data_path: str = "", user_data_path: str = "") -> None:
    """
    Creates the needed folder structure and game config file qrogue_game.config if not already existent.

    :param game_data_path: the path where we want to setup the game data
    :param user_data_path: the path where we want to load and store the user data (e.g. logs, save data)
    :return: None
    """
    Config.setup_user_data(user_data_path)

    path = PathConfig.launch_config_path()
    data = "\n"
    data += f"{game_data_path}\n"
    data += f"{user_data_path}\n"
    PathConfig.write(path, data, in_user_path=False, may_exist=True, append=False)


def start_game(from_console: bool = False, debugging: bool = False, test_level: bool = False,
               data_folder: str = None, user_data_folder: str = None, seed: str = None) -> int:
    if PathConfig.load_paths(data_folder, user_data_folder):
        return_code = Config.load()  # NEEDS TO BE THE FIRST THING WE DO!
    else:
        return_code = 1
    if return_code == 0:
        if debugging:
            Config.activate_debugging(test_level)

        if seed is not None and not debugging:
            print("[Qrogue] Attention! Manual seeds are only available in debug-mode!")
        if seed is None or not debugging:
            seed = random.randint(0, Config.MAX_SEED)
        else:
            seed = int(seed)
        print(f"[Qrogue] Starting game with seed = {seed}")
        try:
            __init_singletons(seed)
            save_data = QrogueCUI(seed).start()
        except PyCuiConfig.OutOfBoundsError:
            # print("[Qrogue] ERROR!")
            # print("Your terminal window is too small. "
            #      "Please make it bigger (i.e. maximize it) or reduce the font size.")
            print("---------------------------------------------------------")
            input("[Qrogue] Press ENTER to close the application")
            sys.exit(1)

        # flush after the player stopped playing
        if GameplayConfig.get_option_value(Options.auto_save):
            success, message = save_data.save()  # todo: save at a different location?
            if success:
                print("[Qrogue] Successfully saved the game!")
            else:
                Logger.instance().error(f"Failed to save the game: {message.text}", show=False, from_pycui=False)
        Logger.instance().flush()
        print("[Qrogue] Successfully flushed all logs and shut down the game without any problems. See you next time!")
    else:
        print(f"[Qrogue] Error #{return_code}:")
        if return_code == 1:
            print("qrogue.config is invalid. Please check if the second line describes a valid path (the path "
                  "to your save files). Using special characters in the path could also cause this error so if the "
                  "path is valid please consider using another one without special characters.")

    if not from_console:
        print()
        input("[Qrogue] Press ENTER to close the application")

    return return_code


def simulate_game(simulation_path: str, from_console: bool = False, debugging: bool = False, data_folder: str = None,
                  user_data_folder: str = None) -> int:
    if PathConfig.load_paths(data_folder, user_data_folder):
        return_code = Config.load()  # NEEDS TO BE THE FIRST THING WE DO!
    else:
        return_code = 1
    if return_code == 0:
        if debugging:
            Config.activate_debugging()

        print(f"[Qrogue] Simulating the game recorded at \"{simulation_path}\"")
        try:
            __init_singletons(seed=PathConfig.get_seed_from_key_log_file(simulation_path))
            save_data = QrogueCUI.start_simulation(simulation_path)
        except PyCuiConfig.OutOfBoundsError:
            # print("[Qrogue] ERROR!")
            # print("Your terminal window is too small. "
            #      "Please make it bigger (i.e. maximize it) or reduce the font size.")
            print("---------------------------------------------------------")
            sys.exit(1)

        if save_data is None:
            print("Simulation failed! Please read the logs to find the reason.")
            Logger.instance().flush()
            sys.exit(1)

        # flush after the player stopped playing
        if GameplayConfig.get_option_value(Options.auto_save):  # todo: save at a different location?
            success, message = save_data.save()
            if success:
                print("[Qrogue] Successfully saved the game!")
            else:
                Logger.instance().error(f"Failed to save the game: {message.text}", show=False, from_pycui=False)
        Logger.instance().flush()
        print("[Qrogue] Successfully flushed all logs and shut down the game without any problems. See you next time!")
    else:
        print(f"[Qrogue] Error #{return_code}:")
        if return_code == 1:
            print("qrogue.config is invalid. Please check if the second line describes a valid path (the path "
                  "to your save files). Using special characters in the path could also cause this error so if the "
                  "path is valid please consider using another one without special characters.")

    if not from_console:
        print()
        input("[Qrogue] Press ENTER to close the application")

    return return_code


def validate_map(path: str, in_base_path: bool = True) -> bool:
    seed = 7
    try:
        Config.load()
        Config.activate_debugging()
        __init_singletons(seed)

        def log_print(title: str, text: str, position: Optional[int] = None):
            # print(f"\"{title}\": {text}")
            pass

        Logger.instance().set_popup(lambda text: log_print(f"Error", text))

    except Exception as ex:
        print(f"Error: {ex}")
        print("Most likely you used validate_map() while also running the game which is forbidden. Make sure to "
              "validate maps separately!")
        return False

    generator = QrogueLevelGenerator(check_achievement_callback=lambda achievement: False,
                                     trigger_event_callback=lambda event: None,
                                     load_map_callback=lambda name, spawn_pos: None,
                                     show_message_callback=lambda title, text, reopen, position: None,
                                     callback_pack=CallbackPack.dummy())

    try:
        error_occurred = False
        generator.generate(seed, file_path=path, in_dungeon_folder=in_base_path)
    except FileNotFoundError as fnf:
        error_occurred = True
        print(f"Could not find specified file! Error: {fnf}")
    except SyntaxError as se:
        error_occurred = True
        print("Found syntax error!")
        print(se)

    return not error_occurred


def play_level(level_name: str, debugging: bool = False, data_folder: str = None, seed: str = None) -> int:
    # NOTE: This does not reset or set any achievements! Therefore, the level might not play the same as if you would
    # play it normally for the first time.
    if PathConfig.load_paths(data_folder, None):
        return_code = Config.load()  # NEEDS TO BE THE FIRST THING WE DO!
    else:
        return_code = 1
    if return_code == 0:
        if seed is not None and not debugging:
            print("[Qrogue] Attention! Manual seeds are only available in debug-mode!")
        if seed is None or not debugging:
            seed = random.randint(0, Config.MAX_SEED)
        else:
            seed = int(seed)
        if debugging: Config.activate_debugging()
        print(f"[Qrogue] Starting level {level_name} with seed = {seed}")
        try:
            __init_singletons(seed)
            QrogueCUI(seed).start(level_name)
        except PyCuiConfig.OutOfBoundsError:
            print("---------------------------------------------------------")
            input("[Qrogue] Press ENTER to close the application")
            sys.exit(1)

        # flush after the player stopped playing
        Logger.instance().flush()
        print("[Qrogue] Successfully flushed all logs and shut down the game without any problems. See you next time!")
    else:
        print(f"[Qrogue] Error #{return_code}:")
        if return_code == 1:
            print("qrogue.config is invalid. Please check if the second line describes a valid path (the path "
                  "to your save files). Using special characters in the path could also cause this error so if the "
                  "path is valid please consider using another one without special characters.")
    return return_code


def update_wfc():
    Logger(print)
    wfc_manager = WFCManager()
    wfc_manager.learn()
    wfc_manager.store()
    Logger.instance().info("WFC matrices updated successfully", from_pycui=False)
    Logger.instance().flush()



from typing import Optional, Callable

from .collectible import Collectible, CollectibleType


class QuantumFuser(Collectible):
    def __init__(self, num: int = 1):
        super().__init__(CollectibleType.QuantumFuser)
        self.__num = num

    def name(self) -> str:
        return "QuantumFuser"

    def description(self, check_unlocks: Optional[Callable[[str], bool]] = None) -> str:
        return f"A {self.name()} allows you to fuse multiple gates into a new one."

    def to_string(self) -> str:
        if self.__num == 1:
            return self.name()
        return f"QuantumFuser #{self.__num}"



from typing import Tuple, Optional

from qrogue.util import Logger
from qrogue.util.config import PathConfig, OptionsManager, Config, ColorConfig, FileTypes
from qrogue.util.controls import Controls, Keys


class GameSimulator:
    # naming convention:
    #   - raw key = PyCUI-internal representation of the physically pressed key
    #   - logical key = QRogue-internal representation of the logically pressed key (e.g., MoveUp, SelectionUp)
    #       ~ multiple logical keys can correspond to the same raw key, both situationally (overlapping controls) and
    #         permanently (same controls)
    #   - key code = "translator" between raw and logical keys
    #       ~ Corresponds to the integer value of a logical key, but as it is computed from raw keys it might not be
    #         the same as its original logical counterpart. However, when given the same Controls-object, it will always
    #         be equivalent to its original logical key.

    __ENCODING = "utf-8"
    __BUFFER_SIZE = 1024
    __VERSION_END = "\n"
    __SEED_END = "\n"
    __SAVE_DATA_START = "Qrogue<"
    __SAVE_DATA_END = ">Qrogue"
    __CONFIG_END = __SAVE_DATA_START

    def __init__(self, path: str, in_keylog_folder: bool = True):
        if not path.endswith(FileTypes.KeyLog.value):
            path += FileTypes.KeyLog.value

        self.__controls: Optional[Controls] = None
        self.__reader = PathConfig.read_keylog_buffered(path, in_keylog_folder, buffer_size=GameSimulator.__BUFFER_SIZE)
        self.__cur_chunk: Optional[bytes] = None
        self._next_chunk()  # initializes __cur_chunk

        self.__notification_popup = True
        self.__marker_counter = 0
        self.__finished = False

        # retrieve the name of the map that was played
        second_line = self.__cur_chunk.index((bytes("\n", GameSimulator.__ENCODING)))
        self.__map_name = str(self.__cur_chunk[0:second_line], GameSimulator.__ENCODING)
        self.__cur_index = second_line

        # change the config so we can reproduce the run (e.g. different auto reset would destroy the simulation)
        if self.__cur_chunk[second_line + 1:].startswith(bytes(Config.HEADER(), GameSimulator.__ENCODING)):
            version_start = second_line + len(Config.HEADER())
            version_end = self.__cur_chunk.index(bytes(self.__VERSION_END, GameSimulator.__ENCODING), version_start)
            self.__version = str(self.__cur_chunk[version_start:version_end], GameSimulator.__ENCODING)

            seed_start = self.__cur_chunk.index(bytes(Config.SEED_HEAD(), GameSimulator.__ENCODING), version_end) \
                         + len(Config.SEED_HEAD())
            seed_end = self.__cur_chunk.index(bytes(self.__SEED_END, GameSimulator.__ENCODING), seed_start)
            self.__seed = int(self.__cur_chunk[seed_start:seed_end])

            time_start = self.__cur_chunk.index(bytes(Config.TIME_HEAD(), GameSimulator.__ENCODING), seed_end) \
                         + len(Config.TIME_HEAD())
            time_end = self.__cur_chunk.index(bytes("\n", GameSimulator.__ENCODING), time_start)
            self.__time = str(self.__cur_chunk[time_start:time_end], GameSimulator.__ENCODING)

            start = self.__cur_chunk.index(bytes(Config.CONFIG_HEAD(), GameSimulator.__ENCODING), seed_end) \
                    + len(Config.CONFIG_HEAD()) + 1  # start at the first line after CONFIG_HEAD
            end = self.__cur_chunk.index(bytes(self.__CONFIG_END, GameSimulator.__ENCODING), start)
            config = str(self.__cur_chunk[start:end], GameSimulator.__ENCODING)
            OptionsManager.from_text(config)

            # todo: use constants instead of these strings
            start = self.__cur_chunk.index(bytes(self.__SAVE_DATA_START, GameSimulator.__ENCODING), end)
            if start < 0:
                self.__save_state = None
            else:
                end = self.__cur_chunk.index(bytes(self.__SAVE_DATA_END, GameSimulator.__ENCODING), start) \
                      + len(self.__SAVE_DATA_END)
                self.__save_state = str(self.__cur_chunk[start:end], GameSimulator.__ENCODING)

            # continue at the \n at the end of the save state because in next_key() we start with going to the next position
            self.__cur_index = end
        else:
            self.__version = "???"
            self.__seed = -1
            self.__time = "???"

    @property
    def map_name(self) -> str:
        return self.__map_name

    @property
    def simulates_over_world(self) -> bool:
        return self.__map_name == "meta"

    @property
    def version(self) -> str:
        return self.__version

    @property
    def seed(self) -> int:
        return self.__seed

    @property
    def time(self) -> str:
        return self.__time

    @property
    def save_state(self) -> Optional[str]:
        return self.__save_state

    def set_controls(self, controls: Controls):
        self.__controls = controls

    def _next_chunk(self):
        chunk = next(self.__reader)
        self.__cur_index = -1
        self.__cur_chunk = bytes(chunk, GameSimulator.__ENCODING)

    def __next_key(self) -> int:
        """

        :return: the next (raw) key or -1 if we should retry (self.__cur_chunk is None if we reached the end)
        """
        self.__cur_index += 1
        if 0 <= self.__cur_index < len(self.__cur_chunk):
            code = self.__cur_chunk[self.__cur_index]
            key_pressed, key = self.__controls.decode(code)
            if key is Keys.ErrorMarker:
                self.__marker_counter += 1
                if self.__marker_counter >= 3:  # todo fix magic number
                    pass  # todo parse error
                return -1
            else:
                self.__marker_counter = 0
            if key_pressed:
                return key_pressed
        else:
            try:
                self._next_chunk()
            except StopIteration:
                self.__cur_chunk = None
        return -1

    def next(self) -> Optional[int]:
        """

        :return: the (raw) key to press or None if the simulation finished
        """
        if self.__notification_popup:
            # close the notification popup if it is still open before using any real simulation keys
            self.__notification_popup = False
            return self.__controls.get_key(Keys.PopupClose)
        while self.__cur_chunk is not None:
            key = self.__next_key()
            if key > -1:
                return key
        self.__finished = True
        return None

    def to_string(self, skip_header: bool = False, skip_raw_keys: bool = False, skip_logical_keys: bool = False) -> str:
        str_repr = ""
        if not skip_header:
            str_repr += self.__version + "\n"
            str_repr += str(self.__seed) + "\n"
            str_repr += str(self.__time) + "\n"
            str_repr += OptionsManager.to_string() + "\n"

        if skip_raw_keys and skip_logical_keys:
            return str_repr

        if not skip_header: str_repr += "\n\nKeys:"  # add keys header

        key_strings = []
        while True:
            k = self.next()
            if k is None: break

            if skip_raw_keys:
                key_strings.append(Keys(self.__controls.encode(k)).name)
            elif skip_logical_keys:
                key_strings.append(str(k))
            else:
                key_strings.append(f"{k} ({Keys(self.__controls.encode(k)).name})")

        str_repr += "\n".join(key_strings) + "\n"
        return str_repr

    def version_warning(self) -> Tuple[str, str]:
        __space = "Space"
        return "Starting Simulation", \
            f"You started a run with \nseed = {self.seed}\nrecorded at {self.time}.\n" \
            f"Press {ColorConfig.highlight_key(__space)} to execute the simulation step by step. " \
            f"Alternatively, if you keep it pressed the simulation will be executed automatically with short " \
            f"delays after each step until you let go of Space again."

    def version_alright(self) -> Tuple[str, str]:
        return "Simulating other version", \
            "You try to simulate the run of a different game version.\n" \
            f"Your current version: {Config.version()}\n" \
            f"Version you try to simulate: {self.version}\n" \
            "This is not supported and can cause problems. Only continue if you know what you do! Else close this " \
            "popup and press ESC to abort the simulation."

    def stop_message(self) -> str:
        if self.__finished:
            text = "finished Simulation"
        else:
            text = "stopped Simulator"
        text += f"\n{Logger.instance().error_count} errors occurred while simulating"
        return text



from antlr4.error.ErrorListener import ErrorListener

from qrogue.util.config import Config


class ParserErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        # print(f"Syntax Error: \"{offendingSymbol}\" at line {line}, column {column} - {msg}")
        raise SyntaxError(msg)

    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
        Config.debug_print("Ambiguity")

    def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
        Config.debug_print("Attempting full context")

    def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
        Config.debug_print("Context sensitivity")



# Generated from D:/Workspaces/pycharm-workspace/Qrogue/qrogue/game/world/dungeon_generator/QrogueDungeon.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .QrogueDungeonParser import QrogueDungeonParser
else:
    from QrogueDungeonParser import QrogueDungeonParser

# This class defines a complete generic visitor for a parse tree produced by QrogueDungeonParser.

class QrogueDungeonVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by QrogueDungeonParser#start.
    def visitStart(self, ctx:QrogueDungeonParser.StartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#meta.
    def visitMeta(self, ctx:QrogueDungeonParser.MetaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#robot.
    def visitRobot(self, ctx:QrogueDungeonParser.RobotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#room_content.
    def visitRoom_content(self, ctx:QrogueDungeonParser.Room_contentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#r_row.
    def visitR_row(self, ctx:QrogueDungeonParser.R_rowContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#tile.
    def visitTile(self, ctx:QrogueDungeonParser.TileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#tile_descriptor.
    def visitTile_descriptor(self, ctx:QrogueDungeonParser.Tile_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#t_descriptor.
    def visitT_descriptor(self, ctx:QrogueDungeonParser.T_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#trigger_descriptor.
    def visitTrigger_descriptor(self, ctx:QrogueDungeonParser.Trigger_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#teleport_descriptor.
    def visitTeleport_descriptor(self, ctx:QrogueDungeonParser.Teleport_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#message_descriptor.
    def visitMessage_descriptor(self, ctx:QrogueDungeonParser.Message_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#collectible_descriptor.
    def visitCollectible_descriptor(self, ctx:QrogueDungeonParser.Collectible_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#energy_descriptor.
    def visitEnergy_descriptor(self, ctx:QrogueDungeonParser.Energy_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#enemy_descriptor.
    def visitEnemy_descriptor(self, ctx:QrogueDungeonParser.Enemy_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#boss_descriptor.
    def visitBoss_descriptor(self, ctx:QrogueDungeonParser.Boss_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#riddle_descriptor.
    def visitRiddle_descriptor(self, ctx:QrogueDungeonParser.Riddle_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#challenge_descriptor.
    def visitChallenge_descriptor(self, ctx:QrogueDungeonParser.Challenge_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#puzzle_parameter.
    def visitPuzzle_parameter(self, ctx:QrogueDungeonParser.Puzzle_parameterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#input_stv.
    def visitInput_stv(self, ctx:QrogueDungeonParser.Input_stvContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#boss_puzzle.
    def visitBoss_puzzle(self, ctx:QrogueDungeonParser.Boss_puzzleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#circuit_stv.
    def visitCircuit_stv(self, ctx:QrogueDungeonParser.Circuit_stvContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#circuit_gate_single.
    def visitCircuit_gate_single(self, ctx:QrogueDungeonParser.Circuit_gate_singleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#circuit_gate_multi.
    def visitCircuit_gate_multi(self, ctx:QrogueDungeonParser.Circuit_gate_multiContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#draw_strategy.
    def visitDraw_strategy(self, ctx:QrogueDungeonParser.Draw_strategyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#stv_pools.
    def visitStv_pools(self, ctx:QrogueDungeonParser.Stv_poolsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#default_stv_pool.
    def visitDefault_stv_pool(self, ctx:QrogueDungeonParser.Default_stv_poolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#stv_pool.
    def visitStv_pool(self, ctx:QrogueDungeonParser.Stv_poolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#stvs.
    def visitStvs(self, ctx:QrogueDungeonParser.StvsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#stv_ref.
    def visitStv_ref(self, ctx:QrogueDungeonParser.Stv_refContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#stv.
    def visitStv(self, ctx:QrogueDungeonParser.StvContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#reward_pools.
    def visitReward_pools(self, ctx:QrogueDungeonParser.Reward_poolsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#default_reward_pool.
    def visitDefault_reward_pool(self, ctx:QrogueDungeonParser.Default_reward_poolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#reward_pool.
    def visitReward_pool(self, ctx:QrogueDungeonParser.Reward_poolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#collectibles.
    def visitCollectibles(self, ctx:QrogueDungeonParser.CollectiblesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#collectible.
    def visitCollectible(self, ctx:QrogueDungeonParser.CollectibleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#integer.
    def visitInteger(self, ctx:QrogueDungeonParser.IntegerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#complex_number.
    def visitComplex_number(self, ctx:QrogueDungeonParser.Complex_numberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#layout.
    def visitLayout(self, ctx:QrogueDungeonParser.LayoutContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#l_room_row.
    def visitL_room_row(self, ctx:QrogueDungeonParser.L_room_rowContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#l_hallway_row.
    def visitL_hallway_row(self, ctx:QrogueDungeonParser.L_hallway_rowContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#rooms.
    def visitRooms(self, ctx:QrogueDungeonParser.RoomsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#room.
    def visitRoom(self, ctx:QrogueDungeonParser.RoomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#r_attributes.
    def visitR_attributes(self, ctx:QrogueDungeonParser.R_attributesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#r_visibility.
    def visitR_visibility(self, ctx:QrogueDungeonParser.R_visibilityContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#r_type.
    def visitR_type(self, ctx:QrogueDungeonParser.R_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#hallways.
    def visitHallways(self, ctx:QrogueDungeonParser.HallwaysContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#hallway.
    def visitHallway(self, ctx:QrogueDungeonParser.HallwayContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#h_attributes.
    def visitH_attributes(self, ctx:QrogueDungeonParser.H_attributesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#messages.
    def visitMessages(self, ctx:QrogueDungeonParser.MessagesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#message.
    def visitMessage(self, ctx:QrogueDungeonParser.MessageContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by QrogueDungeonParser#message_body.
    def visitMessage_body(self, ctx:QrogueDungeonParser.Message_bodyContext):
        return self.visitChildren(ctx)



del QrogueDungeonParser


import enum
import math
from typing import Optional, Tuple, List, Callable, Dict

from qrogue.util.config.py_cui_config import PyCuiColors


class ColorCode(enum.Enum):
    TILE_HIGHLIGHT = "01"
    OBJECT_HIGHLIGHT = "02"
    ACTION_HIGHLIGHT = "03"
    KEY_HIGHLIGHT = "04"
    WORD_HIGHLIGHT = "05"

    INV_TILE_HIGHLIGHT = "11"
    INV_OBJECT_HIGHLIGHT = "12"
    INV_ACTION_HIGHLIGHT = "13"
    INV_KEY_HIGHLIGHT = "14"
    INV_WORD_HIGHLIGHT = "15"

    POPUP_META_INFO = "20"

    PUZZLE_WRONG_AMPLITUDE = "90"
    PUZZLE_CORRECT_AMPLITUDE = "91"
    PUZZLE_HEADLINES = "92"
    PUZZLE_KET = "93"

    def __init__(self, code: str):
        self.__code = code

    def __str__(self):
        return self.__code


class ColorConfig:
    CODE_WIDTH = 2
    SELECTION_COLOR = PyCuiColors.BLACK_ON_WHITE
    QUBIT_INFO_COLOR = PyCuiColors.CYAN_ON_BLACK
    QUBIT_CONFIG_COLOR = PyCuiColors.YELLOW_ON_BLACK
    STV_HEADING_COLOR = PyCuiColors.CYAN_ON_BLACK
    CIRCUIT_COLOR = PyCuiColors.MAGENTA_ON_BLACK
    CIRCUIT_LABEL_COLOR = PyCuiColors.CYAN_ON_BLACK
    QUESTION_SELECTION_COLOR = PyCuiColors.WHITE_ON_BLACK
    LEVEL_SELECTION_HEADER_COLOR = PyCuiColors.CYAN_ON_BLACK
    SEED_INPUT_POPUP_COLOR = PyCuiColors.BLACK_ON_WHITE
    LEVEL_SELECTION_INPUT_MAP_CODE = PyCuiColors.RED_ON_BLACK
    FUSION_CIRCUIT_NAMING_COLOR = PyCuiColors.BLACK_ON_WHITE
    TRANSITION_ACTION_COLOR = PyCuiColors.RED_ON_BLACK

    ERROR_COLOR = PyCuiColors.RED_ON_BLUE
    TEXT_HIGHLIGHT = "//"
    REGEX_TEXT_HIGHLIGHT = "//"  # regex recognizable version of TEXT_HIGHLIGHT (some characters need escaping)
    HIGHLIGHT_WIDTH = len(TEXT_HIGHLIGHT)
    __DIC = {
        str(ColorCode.TILE_HIGHLIGHT):      PyCuiColors.WHITE_ON_BLACK,
        str(ColorCode.OBJECT_HIGHLIGHT):    PyCuiColors.GREEN_ON_WHITE,
        str(ColorCode.ACTION_HIGHLIGHT):    PyCuiColors.RED_ON_WHITE,
        str(ColorCode.KEY_HIGHLIGHT):       PyCuiColors.MAGENTA_ON_WHITE,
        str(ColorCode.WORD_HIGHLIGHT):      PyCuiColors.BLUE_ON_WHITE,

        str(ColorCode.INV_TILE_HIGHLIGHT):      PyCuiColors.BLACK_ON_WHITE,
        str(ColorCode.INV_OBJECT_HIGHLIGHT):    PyCuiColors.GREEN_ON_BLACK,
        str(ColorCode.INV_ACTION_HIGHLIGHT):    PyCuiColors.RED_ON_BLACK,
        str(ColorCode.INV_KEY_HIGHLIGHT):       PyCuiColors.MAGENTA_ON_BLACK,
        str(ColorCode.INV_WORD_HIGHLIGHT):      PyCuiColors.BLUE_ON_BLACK,

        str(ColorCode.POPUP_META_INFO):         PyCuiColors.WHITE_ON_MAGENTA,

        str(ColorCode.PUZZLE_WRONG_AMPLITUDE):      PyCuiColors.RED_ON_BLACK,
        str(ColorCode.PUZZLE_CORRECT_AMPLITUDE):    PyCuiColors.GREEN_ON_BLACK,
        str(ColorCode.PUZZLE_HEADLINES):            STV_HEADING_COLOR,
        str(ColorCode.PUZZLE_KET):                  QUBIT_CONFIG_COLOR,
    }

    @staticmethod
    def is_number(text: str) -> bool:
        try:
            int(text)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_punctuation(char: str) -> bool:
        return char in [".", ",", "!", "?", ":", "\""]

    @staticmethod
    def __find(paragraph: str, start: int, end: int) -> int:
        # adapt end because meta characters are not printed and therefore they can be directly after end
        end += ColorConfig.HIGHLIGHT_WIDTH - 1
        return paragraph.find(ColorConfig.TEXT_HIGHLIGHT, start, end)

    @staticmethod
    def count_meta_characters(paragraph: str, width: int, handle_error: Callable[[str], None]) -> int:
        """
        Counts how many meta characters (i.e. not printed characters) there are in the first #width characters of
        paragraph. This way we know for example by how much we can extend the rendered text since these characters
        won't be rendered.

        :param paragraph: the str we won't to count the number of meta characters for
        :param width: number of characters we consider in paragraph (i.e. line width)
        :param handle_error: handles potential errors
        :return: number of found meta characters
        """
        character_removals = 0
        # check how many meta-characters (indicating color rules) we have in our line
        highlight_index = ColorConfig.__find(paragraph, 0, width)
        start = True  # whether we search for the start of a highlighted section or an end
        while highlight_index > -1:
            highlight_index += ColorConfig.HIGHLIGHT_WIDTH
            if start:
                if highlight_index + ColorConfig.CODE_WIDTH <= len(paragraph) and \
                        ColorConfig.is_number(paragraph[highlight_index:highlight_index + ColorConfig.CODE_WIDTH]):
                    last_whitespace = paragraph.rfind(" ", highlight_index,
                                                      width + character_removals + 1 + ColorConfig.CODE_WIDTH)
                    if last_whitespace > -1:
                        character_removals += ColorConfig.HIGHLIGHT_WIDTH + ColorConfig.CODE_WIDTH
                        start = False
                    elif paragraph.endswith(ColorConfig.TEXT_HIGHLIGHT):
                        # due to splitting a line in the middle of a color rule it can happen that there is no " "
                        # at the end but a "/" and therefore it would still fit
                        character_removals += ColorConfig.HIGHLIGHT_WIDTH + ColorConfig.CODE_WIDTH \
                                              + ColorConfig.HIGHLIGHT_WIDTH
                        break
                    elif ColorConfig.is_punctuation(paragraph[-1]):
                        # TODO I don't think -1 is correct, because it is the very end, but somehow it works
                        # if the very last word is highlighted we also have no whitespace at the end
                        character_removals += ColorConfig.HIGHLIGHT_WIDTH + ColorConfig.CODE_WIDTH \
                                              + ColorConfig.HIGHLIGHT_WIDTH
                        break
                    else:
                        break
                else:
                    handle_error(f"Illegal start index = {highlight_index} for \"{paragraph}\". Make sure no text"
                                 f" contains \"{ColorConfig.TEXT_HIGHLIGHT}\" or a 2 or more digit number directly"
                                 f" after a highlighting (space in-between is okay)!")
            else:
                character_removals += ColorConfig.HIGHLIGHT_WIDTH
                start = True
            highlight_index = ColorConfig.__find(paragraph, highlight_index, width + character_removals)
        return character_removals

    @staticmethod
    def colorize(color_code: ColorCode, text) -> str:
        return f"{ColorConfig.TEXT_HIGHLIGHT}{color_code}{text}{ColorConfig.TEXT_HIGHLIGHT}"

    @staticmethod
    def highlight_tile(tile: str, invert: Optional[bool] = None) -> str:
        """
        Highlights tile strings.
        :param tile: a string to highlight as tile
        :param invert: whether to use the normal (default) or inverted background colors
        :return:
        """
        if invert is None: invert = False
        return ColorConfig.colorize(ColorCode.INV_TILE_HIGHLIGHT if invert else ColorCode.TILE_HIGHLIGHT, tile)

    @staticmethod
    def ht(tile: str, invert: Optional[bool] = None) -> str:
        """
        Shortcut for ColorConfig.highlight_tile().
        """
        return ColorConfig.highlight_tile(tile, invert)

    @staticmethod
    def highlight_object(obj: str, invert: Optional[bool] = None) -> str:
        """
        Highlights something directly gameplay related. I.e., things you encounter in the game.
        :param obj: a string to highlight as object
        :param invert: whether to use the normal (default) or inverted background colors
        :return:
        """
        if invert is None: invert = False
        return ColorConfig.colorize(ColorCode.INV_OBJECT_HIGHLIGHT if invert else ColorCode.OBJECT_HIGHLIGHT, obj)

    @staticmethod
    def ho(obj: str, invert: Optional[bool] = None) -> str:
        """
        Shortcut for ColorConfig.highlight_object().
        """
        return ColorConfig.highlight_object(obj, invert)

    @staticmethod
    def highlight_action(action: str, invert: Optional[bool] = None) -> str:
        """
        Highlights action words that explain what you can do in the game.
        :param action: a string to highlight as action
        :param invert: whether to use the normal (default) or inverted background colors
        :return:
        """
        if invert is None: invert = False
        return ColorConfig.colorize(ColorCode.INV_ACTION_HIGHLIGHT if invert else ColorCode.ACTION_HIGHLIGHT, action)

    @staticmethod
    def ha(action: str, invert: Optional[bool] = None) -> str:
        """
        Shortcut for ColorConfig.highlight_action().
        """
        return ColorConfig.highlight_action(action, invert)

    @staticmethod
    def highlight_key(key: str, invert: Optional[bool] = None) -> str:
        """
        Highlights a keyboard input.
        :param key: a string to highlight as key
        :param invert: whether to use the normal (default) or inverted background colors
        :return:
        """
        if invert is None: invert = False
        return ColorConfig.colorize(ColorCode.INV_KEY_HIGHLIGHT if invert else ColorCode.KEY_HIGHLIGHT, key)

    @staticmethod
    def hk(key: str, invert: Optional[bool] = None) -> str:
        """
        Shortcut for ColorConfig.highlight_key().
        """
        return ColorConfig.highlight_key(key, invert)

    @staticmethod
    def highlight_word(word: str, invert: Optional[bool] = None) -> str:
        """
        Highlights miscellaneous words that should be highlighted.
        :param word: a string to highlight as important word
        :param invert: whether to use the normal (default) or inverted background colors
        :return:
        """
        if invert is None: invert = False
        return ColorConfig.colorize(ColorCode.INV_WORD_HIGHLIGHT if invert else ColorCode.WORD_HIGHLIGHT, word)

    @staticmethod
    def hw(word: str, invert: Optional[bool] = None) -> str:
        """
        Shortcut for ColorConfig.highlight_word().
        """
        return ColorConfig.highlight_word(word, invert)

    @staticmethod
    def get(char: str) -> int:
        try:
            return ColorConfig.__DIC[char]
        except KeyError:
            return ColorConfig.ERROR_COLOR

    @staticmethod
    def get_from_code(code: ColorCode) -> int:
        return ColorConfig.get(str(code))


class PopupConfig:
    class Importance(enum.IntEnum):
        Undefined = 0
        Dialogue = 1
        Info = 2
        Error = 3

        @staticmethod
        def from_int(value: int) -> "PopupConfig.Importance":
            if value == PopupConfig.Importance.Dialogue: return PopupConfig.Importance.Dialogue
            if value == PopupConfig.Importance.Info: return PopupConfig.Importance.Info
            if value == PopupConfig.Importance.Error: return PopupConfig.Importance.Error
            return PopupConfig.Importance.Undefined

    # note that the metric is:
    # 0=Center, >0 = distance to left or top border, <0 = distance to right or bottom border
    __DEFAULT_POSITION = 0
    __CENTER_POS = 0    # special number to indicate the center
    __TOP_POS = 3       # three rows below top
    __RIGHT_POS = -5    # five columns left of right
    __BOT_POS = -3      # three rows above bot
    __LEFT_POS = 5      # five columns right of left
    # just a very large number that we definitely won't use for regular positions to indicate +0 and -0:
    __ZERO_POS = 1_000_000
    __POSITIONS = [  # None = center
        (__CENTER_POS, __CENTER_POS), (__CENTER_POS, __TOP_POS), (__RIGHT_POS, __TOP_POS), (__RIGHT_POS, __CENTER_POS),
        (__RIGHT_POS, __BOT_POS), (__CENTER_POS, __BOT_POS), (__LEFT_POS, __BOT_POS), (__LEFT_POS, __CENTER_POS),
        (__LEFT_POS, __TOP_POS),
        # special handcrafted positions
        (+__ZERO_POS, 6),  # for matrix popup on top of normal matrix widget
    ]
    __INDEX_X = 0  # position of X in position tuple
    __INDEX_Y = 1  # position of Y in position tuple

    PADDING_X = 2
    PADDING_Y = 2

    @staticmethod
    def scroll_indicator() -> str:
        return "scroll down"

    @staticmethod
    def default_color() -> int:
        return PyCuiColors.BLACK_ON_WHITE

    @staticmethod
    def default_position() -> int:
        assert 0 <= PopupConfig.__DEFAULT_POSITION < len(PopupConfig.__POSITIONS), "Invalid DefaultPosition for Popups!"
        return PopupConfig.__DEFAULT_POSITION

    @staticmethod
    def __resolve_pos_val(pos: int, val_index: int) -> Tuple[Optional[bool], int]:
        """

        Args:
            pos: position indicated via index in internal __POSITIONS list
            val_index: index of the value in the position tuple

        Returns: A tuple indicating the reference position (left or up=True, right or bottom=False, center=None) and
        the distance of the indexed value to its reference position. E.g., a position near the bottom right corner could
        return (False, 3), meaning its either 3 rows (if val_index corresponds to y-position) above the bottom of the
        window or 3 columns (if val_index corresponds to x-position) to the left of the right end of the window.

        """
        if 0 <= pos < len(PopupConfig.__POSITIONS):
            val = PopupConfig.__POSITIONS[pos][val_index]
            if val == +PopupConfig.__ZERO_POS: return True, 0
            if val == -PopupConfig.__ZERO_POS: return False, 0
            if val > 0: return True, val
            if val < 0: return False, -val  # since we return a distance, the value needs to be positive
            return None, 0  # val must be "normal" 0 (or None corresponding to 0)
        else:
            return PopupConfig.__resolve_pos_val(PopupConfig.default_position(), val_index)

    @staticmethod
    def resolve_position_x(pos: int) -> Tuple[Optional[bool], int]:
        """

        Args:
            pos: position indicated via index in internal __POSITIONS list

        Returns: A tuple indicating the reference position (left=True, right=False, center=None) and the x-distance to
        the reference position. E.g., a position near the bottom right corner could return (False, 3), meaning its 3
        columns to the left of the right end of the window.

        """
        return PopupConfig.__resolve_pos_val(pos, PopupConfig.__INDEX_X)

    @staticmethod
    def resolve_position_y(pos: int) -> Tuple[Optional[bool], int]:
        """

        Args:
            pos: position indicated via index in internal __POSITIONS list

        Returns: A tuple indicating the reference position (top=True, bottom=False, center=None) and the y-distance to
        the reference position. E.g., a position near the bottom right corner could return (False, 3), meaning its 3
        rows above the bottom of the window.

        """
        return PopupConfig.__resolve_pos_val(pos, PopupConfig.__INDEX_Y)


class UIConfig:
    class Dimensions:
        @staticmethod
        def __from_margins(margin_left: int, margin_top: int, center_x: bool = True, center_y: bool = False,
                           pad_x: Optional[int] = None, pad_y: Optional[int] = None):
            # NOT FULLY WORKING I THINK
            if margin_left < 0 and not center_x:
                width = UIConfig.WINDOW_WIDTH + margin_left
                margin_left = 0     # it's 0 since we place it from the right
            else:
                # if margin_left > 0: nothing changes; else we want to center anyways and can place from the left
                margin_left = abs(margin_left)
                width = UIConfig.WINDOW_WIDTH - margin_left
                if center_x: width -= margin_left  # subtract an additional time for the other side

            if margin_top < 0 and not center_y:
                height = UIConfig.WINDOW_HEIGHT + margin_top
                margin_top = 0
            else:
                abs_margin_top = abs(margin_top)
                height = UIConfig.WINDOW_HEIGHT-abs_margin_top
                if center_y: height -= abs_margin_top   # subtract an additional time for the other side

            return UIConfig.Dimensions(width, height, margin_left, margin_top, pad_x, pad_y)

        def __init__(self, width: int, height: int, margin_left: Optional[int] = None, margin_top: Optional[int] = None,
                     pad_x: Optional[int] = None, pad_y: Optional[int] = None):
            # check for 0 and relative 0 if using negative values
            assert width not in (0, -UIConfig.WINDOW_WIDTH), f"Width ({width}) must not be equivalent to 0!"
            assert height not in (0, -UIConfig.WINDOW_HEIGHT), f"Height ({height}) must not be equivalent to 0!"
            assert abs(width) <= UIConfig.WINDOW_WIDTH, f"Width ({width}) is out of bounds!"
            assert abs(height) <= UIConfig.WINDOW_HEIGHT, f"Height ({height}) is out of bounds!"

            # negative values are interpreted as relative to max dimensions
            if width < 0: width = UIConfig.WINDOW_WIDTH + width
            if height < 0: height = UIConfig.WINDOW_HEIGHT + height

            if margin_left is None: margin_left = 0
            if margin_top is None: margin_top = 0

            # negative margins are interpreted as absolute values from right instead of left or bottom instead of top
            if margin_left < 0: margin_left = UIConfig.WINDOW_WIDTH + margin_left
            if margin_top < 0: margin_top = UIConfig.WINDOW_HEIGHT + margin_top

            assert width + margin_left <= UIConfig.WINDOW_WIDTH, f"Invalid width or margin_left: {width} + " \
                                                                 f"{margin_left} <= {UIConfig.WINDOW_WIDTH} is False!"
            assert height + margin_top <= UIConfig.WINDOW_HEIGHT, f"Invalid height or margin_top: {height} + " \
                                                                  f"{margin_top} <= {UIConfig.WINDOW_HEIGHT} is False!"

            # default values used from WidgetSet.add_block_label()
            if pad_x is None: pad_x = 1
            if pad_y is None: pad_y = 0

            self.__width = width
            self.__height = height
            self.__margin_left = margin_left
            self.__margin_top = margin_top
            self.__pad_x = pad_x
            self.__pad_y = pad_y

        @property
        def width(self) -> int:
            return self.__width

        @property
        def height(self) -> int:
            return self.__height

        @property
        def margin_left(self) -> int:
            return self.__margin_left

        @property
        def margin_top(self) -> int:
            return self.__margin_top

        @property
        def pad_x(self) -> int:
            return self.__pad_x

        @property
        def pad_y(self) -> int:
            return self.__pad_y

        # implied properties
        @property
        def left(self) -> int:
            return self.__margin_left

        @property
        def top(self) -> int:
            return self.__margin_top

        @property
        def right(self) -> int:
            return self.__margin_left + self.__width

        @property
        def bottom(self) -> int:
            return self.__margin_top + self.__height

        def center(self, horizontal: bool = False, vertical: bool = False) -> "UIConfig.Dimensions":
            width, height, margin_left, margin_top, pad_x, pad_y = self
            if horizontal:
                margin_left = int((UIConfig.WINDOW_WIDTH - width) / 2)
            if vertical:
                margin_top = int((UIConfig.WINDOW_HEIGHT - height) / 2)
            return UIConfig.Dimensions(width, height, margin_left, margin_top, pad_x, pad_y)

        def align_horizontal(self, left: bool) -> "UIConfig.Dimensions":
            width, height, _, margin_top, pad_x, pad_y = self
            margin_left = 0 if left else UIConfig.WINDOW_WIDTH - width
            return UIConfig.Dimensions(width, height, margin_left, margin_top, pad_x, pad_y)

        def align_vertical(self, top: bool) -> "UIConfig.Dimensions":
            width, height, margin_left, _, pad_x, pad_y = self
            margin_top = 0 if top else UIConfig.WINDOW_HEIGHT - height
            return UIConfig.Dimensions(width, height, margin_left, margin_top, pad_x, pad_y)

        def from_base(self, new_values: Dict[str, int]):
            width, height, margin_left, margin_top, pad_x, pad_y = self
            if 'width' in new_values: width = new_values['width']
            if 'height' in new_values: height = new_values['height']

            # allow aliases for margins
            if 'margin_left' in new_values: margin_left = new_values['margin_left']
            if 'margin_x' in new_values: margin_left = new_values['margin_x']
            if 'margin_top' in new_values: margin_top = new_values['margin_top']
            if 'margin_y' in new_values: margin_top = new_values['margin_y']

            if 'pad_x' in new_values: pad_x = new_values['pad_x']
            if 'pad_y' in new_values: pad_y = new_values['pad_y']
            return UIConfig.Dimensions(width, height, margin_left, margin_top, pad_x, pad_y)

        def relative(self, add_values: Dict[str, int]):
            width, height, margin_left, margin_top, pad_x, pad_y = self
            if 'width' in add_values: width += add_values['width']
            if 'height' in add_values: height += add_values['height']

            # allow aliases for margins
            if 'margin_left' in add_values: margin_left += add_values['margin_left']
            if 'margin_x' in add_values: margin_left += add_values['margin_x']
            if 'margin_top' in add_values: margin_top += add_values['margin_top']
            if 'margin_y' in add_values: margin_top += add_values['margin_y']

            if 'pad_x' in add_values: pad_x += add_values['pad_x']
            if 'pad_y' in add_values: pad_y += add_values['pad_y']
            return UIConfig.Dimensions(width, height, margin_left, margin_top, pad_x, pad_y)

        def __iter__(self):
            return iter(
                (self.__width, self.__height, self.__margin_left, self.__margin_top, self.__pad_x, self.__pad_y)
            )

        def __str__(self):
            return f"WS-Dims(size={self.__width}x{self.__height}|margin={self.__margin_left},{self.__margin_top}|" \
                   f"pad:{self.__pad_x},{self.__pad_y})"

    # measures are in PyCUI rows and columns, not characters!
    WINDOW_WIDTH = 17
    WINDOW_HEIGHT = 10
    WINDOW_DIMS: Optional[Dimensions] = None

    HUD_HEIGHT = 1
    HUD_WIDTH = math.floor(WINDOW_WIDTH / 2)
    NON_HUD_HEIGHT = WINDOW_HEIGHT - HUD_HEIGHT

    PAUSE_CHOICES_WIDTH = math.floor(WINDOW_WIDTH / 3)
    PAUSE_DESCRIPTION_HEIGHT = math.floor((WINDOW_HEIGHT - HUD_HEIGHT) / 3)

    MAIN_MENU_ROW = 2
    MAIN_MENU_HEIGHT = round(WINDOW_HEIGHT / 2)
    ASCII_ART_WIDTH = math.floor(2 * WINDOW_WIDTH / 3)

    TRANSITION_SCREEN_HEIGHT = 3
    TRANSITION_SCREEN_WIDTH = 5
    TRANSITION_SCREEN_ROW = 3
    TRANSITION_SCREEN_COL = int((WINDOW_WIDTH - TRANSITION_SCREEN_WIDTH) / 2)

    INPUT_STV_WIDTH = 2
    OUTPUT_STV_WIDTH = 2
    TARGET_STV_WIDTH = 3
    STV_HEIGHT = math.floor(WINDOW_HEIGHT * 0.6)
    DIALOG_HEIGHT = 2

    LEVEL_SELECT_SUMMARY_X = 4
    LEVEL_SELECT_SUMMARY_Y = 1
    LEVEL_SELECT_SUMMARY_SEED_WIDTH = 2
    LEVEL_SELECT_MAIN_X = 1
    LEVEL_SELECT_MAIN_Y = 2
    LEVEL_SELECT_MAIN_HEIGHT = 5
    LEVEL_SELECT_CHOICES_WIDTH = 3
    LEVEL_SELECT_DETAILS_WIDTH = 7

    WB_MARGIN_X = 1
    WB_ACTIONS_WIDTH = 2
    WB_INFO_HEIGHT = 2

    WB_RESOURCES_DIMS: Optional[Dimensions] = None
    WB_ACTIONS_DIMS: Optional[Dimensions] = None
    WB_GATES_DIMS: Optional[Dimensions] = None
    WB_INFOS_DIMS: Optional[Dimensions] = None

    @staticmethod
    def init():
        UIConfig.WINDOW_DIMS = UIConfig.Dimensions(width=UIConfig.WINDOW_WIDTH, height=UIConfig.WINDOW_HEIGHT)
        UIConfig._init_workbench()

    @staticmethod
    def _init_workbench():
        # first store the dimensions in local variables for better readability
        global_margin_x = 1
        res = UIConfig.Dimensions(-2*global_margin_x, 2).center(horizontal=True).align_vertical(top=True)
        infos = UIConfig.Dimensions(-2*global_margin_x, 2).center(horizontal=True).align_vertical(top=False)

        # fill actions and gates height-wise in between res and infos
        actions = UIConfig.Dimensions(3, -(res.height + infos.height), global_margin_x, res.bottom)
        gates = UIConfig.Dimensions(-(actions.right + global_margin_x), actions.height, actions.right,
                                    actions.margin_top)

        # now set the values of the global variables
        UIConfig.WB_RESOURCES_DIMS = res
        UIConfig.WB_ACTIONS_DIMS = actions
        UIConfig.WB_GATES_DIMS = gates
        UIConfig.WB_INFOS_DIMS = infos

    @staticmethod
    def wb_info_dims() -> Dimensions:
        return UIConfig.Dimensions(-2, 2).center(horizontal=True).align_vertical(top=False)
        #return UIConfig.Dimensions(UIConfig.WINDOW_WIDTH-2, 2, 1, -2)
        #return UIConfig.Dimensions.from_margins(1, -2)

    @staticmethod
    def stv_height(num_of_qubits: int) -> int:
        if num_of_qubits == 1:
            return 1
        elif num_of_qubits == 2:
            return 2
        elif num_of_qubits == 3:
            return 4
        else:
            return 6


class HudConfig:
    ShowScore = True
    ShowMapName = True
    ShowEnergy = False
    ShowKeys = True
    ShowFPS = False


# located here and not in util_functions.py because it depends on visual configs
def split_text(text: str, width: int, padding: int, handle_error: Callable[[str], None]) -> List[str]:
    """

    :param text: text to split into multiple lines
    :param width: the maximum width of one line
    :param padding: how much spaces we should place at the start and end of each line
    :param handle_error: to handle potential errors
    :return: list of text parts with a maximum of #width characters, every text part is rstrip()-ed     # todo: test thoroughly
    """
    width -= padding * 2  # remove the space needed for padding
    lines = []
    for paragraph in text.splitlines():
        paragraph = paragraph.rstrip()  # remove any redundant whitespace after the paragraph
        index = 0
        prepend: Optional[str] = None
        while True:  # technically: while index < len(paragraph), but we can only exit the loop with our break
            cur_part = paragraph[index:]

            # the beginning of a paragraph can have intentional whitespace for indentation, created lines are stripped
            if index > 0 and cur_part.startswith(" "):
                temp = cur_part.lstrip()
                # since we removed whitespace, we need to advance index so cur_part == paragraph[index:] is True again
                index += len(cur_part) - len(temp)  # the length difference tells us how many characters we removed
                cur_part = temp

            # check if the previous line ended with a color rule and prepend it
            if prepend is not None:
                cur_part = prepend + cur_part
                # we have to adapt the index since we potentially remove whitespace in front and therefore have
                # the possibility of a longer line which leads to a higher increment of the index and furthermore
                # to the potential loss of some characters in the beginning of the new_line
                # prepend also needs to be part of this adaption because it will be counted in character_removals
                index -= len(prepend)
                prepend = None

            # the currently usable width is the displayable width + the number of occurring non-printable characters
            cur_width = width + ColorConfig.count_meta_characters(cur_part, width, handle_error)

            # if cur_part fits we can skip the next steps and just append it as new line (after while, in for)
            if len(cur_part) <= cur_width:
                last_line = cur_part
                break

            # Split the line at its last reachable whitespace.
            # +1 because if the last whitespace is exactly after cur_width characters, the line still fits and the
            # additional character we get in the slice will be stripped anyways (because it's the found whitespace).
            # Since cur_part is stripped, we could start at 1, but don't due it for readability-reasons.
            last_whitespace = cur_part.rfind(" ", 0, cur_width + 1)
            if last_whitespace >= 0: cur_width = last_whitespace + 1

            next_line = cur_part[:cur_width].rstrip()
            if len(next_line) > 0:
                # check if next_line ends with an un-terminated color rule
                last_highlight = next_line.rfind(ColorConfig.TEXT_HIGHLIGHT)
                if 0 <= last_highlight < len(next_line) - ColorConfig.HIGHLIGHT_WIDTH:
                    # if so, terminate it and remember to continue it in the next line
                    code_start = last_highlight + ColorConfig.HIGHLIGHT_WIDTH
                    # TODO a highlighted number can potentially lead to problems here! (at least theoretically,
                    # todo but somehow I couldn't produce a breaking example so I might be wrong)
                    # todo also if we place an at-least-2-digits number directly after a highlight
                    if ColorConfig.is_number(next_line[code_start:code_start + ColorConfig.CODE_WIDTH]):
                        next_line += ColorConfig.TEXT_HIGHLIGHT
                        prepend = next_line[last_highlight:code_start + ColorConfig.CODE_WIDTH]
                lines.append(next_line)
            index += cur_width  # index will never increase to >= len(paragraph)

        # we don't have anything to prepend since we already added it before breaking the loop
        lines.append(last_line)

    return [" " * padding + line + " " * padding for line in lines]



from typing import Callable, Tuple, List, Optional

from py_cui import ColorRule
from py_cui.popups import Popup as PyCuiPopup
from py_cui.ui import MenuImplementation

from qrogue.util import ColorConfig as CC, Keys, Logger, PopupConfig, ColorConfig, ColorCode, split_text, Config


class MultilinePopup(PyCuiPopup, MenuImplementation):
    __STATIC_PY_CUI_PADDING = 6  # based on PyCUI "padding" I think
    __QUESTION_SPACING = " " * 7
    __QUESTION_ARROW = "-->"
    __QUESTION_SELECTION = " " * (len(__QUESTION_SPACING) - (len(__QUESTION_ARROW) + 1)) + (__QUESTION_ARROW + " ")
    __BASE_DIMENSIONS = 11, 100

    @staticmethod
    def popup_width_to_content_width(popup_width: int, padding: int = 0) -> int:
        return popup_width - MultilinePopup.__STATIC_PY_CUI_PADDING - 2 * padding

    @staticmethod
    def __get_color_rules():
        regex = CC.REGEX_TEXT_HIGHLIGHT
        return [
            ColorRule(f"{regex}.*?{regex}", 0, 0, "contains", "regex", [0, 1],
                      False, Logger.instance()),
            ColorRule(MultilinePopup.__QUESTION_ARROW, ColorConfig.QUESTION_SELECTION_COLOR, 0, "contains", "regex",
                      [0, 1], False, Logger.instance()),
        ]

    def __init__(self, root, title, text, color, renderer, logger: Logger, controls,
                 confirmation_callback: Callable[[int], None] = None, answers: Optional[List] = None,
                 pos: Optional[int] = None, dimensions: Optional[Tuple[int, int]] = None,
                 situational_callback: Optional[Tuple[Optional[Callable[[], None]], Optional[Callable[[], None]]]] =
                 None, padding_x: Optional[int] = None):
        """
        :param dimensions: expected size of the popup's content. The popup itself will be a bit bigger.
        """
        # custom_size needs to be initialized immediately because get_absolute_stop_pos() in init() accesses it
        # we would somehow get stuck in an endless loop if we would manually calculate dimensions, starts and stops
        #  before init() hence we use this workaround
        self.__custom_size = False

        if padding_x is None: padding_x = PopupConfig.PADDING_X

        super().__init__(root, title, text, color, renderer, logger)
        self.__controls = controls
        self.__confirmation_callback = confirmation_callback
        if answers is None and self._is_question:
            self.__answers = ["Confirm", "Cancel"]
        else:
            self.__answers = answers

        self.__custom_pos = pos
        self.__custom_size = True
        if dimensions is None:
            self._height, self._width = MultilinePopup.__BASE_DIMENSIONS
        else:
            # custom dimensions state the size of the content of the popup, so we add additional paddings
            self._height = dimensions[0] + 2  # top and bottom row are extra
            # inverse of popup_width_to_content_width():
            self._width = dimensions[1] + MultilinePopup.__STATIC_PY_CUI_PADDING + 2 * padding_x
        self._start_x, self._start_y = self.get_absolute_start_pos()
        self._stop_x, self._stop_y = self._start_x + self._width, self._start_y + self._height

        def _dummy():
            pass

        self.__situational_callback1 = _dummy
        self.__situational_callback2 = _dummy
        if situational_callback is not None:
            sit1, sit2 = situational_callback
            if sit1 is not None: self.__situational_callback1 = sit1
            if sit2 is not None: self.__situational_callback2 = sit2

        self._top_view = 0
        self.__lines = split_text(text, self._width - MultilinePopup.__STATIC_PY_CUI_PADDING, padding_x,
                                  lambda err: logger.error(err, show=False, from_pycui=False))

        self.__question_state = 0
        self._pageAlignment = " " * (self._width - MultilinePopup.__STATIC_PY_CUI_PADDING)

        if Config.debugging(): self.set_title(self.get_title() + f" {self.get_absolute_dimensions()}")

    @property
    def _is_question(self) -> bool:
        return self.__confirmation_callback is not None

    @property
    def textbox_height(self) -> int:
        """

        :return: number of rows inside the popup
        """
        full_page_height = self._height - self._pady - 2  # subtract both title and end line
        if self._is_question:
            full_page_height -= 2  # subtract the two lines needed for answer selection
        if len(self.__lines) > full_page_height:
            return full_page_height - 1  # subtract the space we need for page indication
        return full_page_height

    def __up(self):
        if len(self.__lines) > self.textbox_height and self._top_view > 0:
            self._top_view -= 1

    def __down_fast(self):
        if len(self.__lines) > self.textbox_height and self._top_view > 0:
            self._top_view = max(self._top_view - self.textbox_height, 0)

    def __down(self):
        # since _top_view can never be negative we don't need to check if we have lines that don't fit on the popup
        if self._top_view < len(self.__lines) - self.textbox_height:
            self._top_view += 1

    def __up_fast(self):
        if self._top_view < len(self.__lines) - self.textbox_height:
            self._top_view = min(self._top_view + self.textbox_height, len(self.__lines) - self.textbox_height)

    def _handle_key_press(self, key_pressed):
        """Overrides base class handle_key_press function
        """
        # regarding the idea to move this functionality to Keys.PopupLeft and Keys.PopupRight:
        # these keys should definitely be used for selecting the answer in question popups, so it would only make it
        # more inconsistent and confusing im question popups and classical popups have different keys for, e.g.,
        # navigating through history
        if key_pressed in self.__controls.get_keys(Keys.Situational1):
            self.__situational_callback1()
        elif key_pressed in self.__controls.get_keys(Keys.Situational2):
            self.__situational_callback2()

        elif self._is_question:
            if key_pressed in self.__controls.get_keys(Keys.Action):
                self.__confirmation_callback(self.__question_state)
                self._root.close_popup()
            elif key_pressed in self.__controls.get_keys(Keys.PopupRight):
                self.__question_state = min(self.__question_state + 1, len(self.__answers) - 1)
            elif key_pressed in self.__controls.get_keys(Keys.PopupLeft):
                self.__question_state = max(self.__question_state - 1, 0)
            elif key_pressed in self.__controls.get_keys(Keys.PopupUp):
                self.__up()
            elif key_pressed in self.__controls.get_keys(Keys.PopupDown):
                self.__down()
        else:
            if key_pressed in self.__controls.get_keys(Keys.PopupClose):
                self._root.close_popup()
            elif key_pressed in self.__controls.get_keys(Keys.PopupUp):
                self.__up()
            elif key_pressed in self.__controls.get_keys(Keys.PopupLeft):
                self.__down_fast()
            elif key_pressed in self.__controls.get_keys(Keys.PopupDown):
                self.__down()
            elif key_pressed in self.__controls.get_keys(Keys.PopupRight):
                self.__up_fast()

    def _draw(self):
        """Overrides base class draw function
        """

        self._renderer.set_color_mode(self._color)
        self._renderer.draw_border(self)
        self._renderer.set_color_rules(self.__get_color_rules())
        counter = self._pady + 1
        for i in range(self._top_view, len(self.__lines)):
            if counter > self.textbox_height:
                break
            self._renderer.draw_text(self, self.__lines[i], self._start_y + counter, selected=True)
            counter += 1
            i += 1

        info_line_y = self._start_y + self.textbox_height + 1
        if len(self.__lines) > self.textbox_height:
            rem_rows = str(len(self.__lines) - counter - self._top_view + 1)
            rem_rows = self._pageAlignment[:-len(rem_rows)] + ColorConfig.colorize(ColorCode.POPUP_META_INFO, rem_rows)
            prefix = PopupConfig.scroll_indicator()
            rem_rows = ColorConfig.colorize(ColorCode.POPUP_META_INFO, prefix) + rem_rows[len(prefix):]
            self._renderer.draw_text(self, rem_rows, info_line_y, selected=False)
            info_line_y += 1

        if self._is_question:
            # the question answers should always be positioned at the very bottom of the popup
            self._renderer.draw_text(self, "-" * self._width, info_line_y, selected=True)

            text = ""
            for i in range(len(self.__answers)):
                if i == self.__question_state:
                    text += MultilinePopup.__QUESTION_SELECTION
                else:
                    text += MultilinePopup.__QUESTION_SPACING
                text += self.__answers[i]
            self._renderer.draw_text(self, text, info_line_y + 1, selected=False)

        self._renderer.unset_color_mode(self._color)
        self._renderer.reset_cursor(self)

    def get_absolute_start_pos(self) -> Tuple[int, int]:
        if self.__custom_size:
            cui_height, cui_width = self._root.get_absolute_size()

            # start relative to the middle
            start_x = int((cui_width - self._width) / 2)
            start_y = int((cui_height - self._height) / 2)

            # adapt start_x and start_y for custom positions
            if self.__custom_pos is not None:
                # reference position (left, right or center) and value
                ref_x, val_x = PopupConfig.resolve_position_x(self.__custom_pos)
                if ref_x is True:
                    start_x = val_x
                elif ref_x is False:
                    start_x = (cui_width - self._width) - val_x
                # else ref_x is None, indicating center, so we don't have to adjust start_x

                ref_y, val_y = PopupConfig.resolve_position_y(self.__custom_pos)
                if ref_y is True:
                    start_y = val_y
                elif ref_y is False:
                    start_y = (cui_height - self._height) - val_y
                # else ref_y is None, indicating center, so we don't have to adjust start_y

            return start_x, start_y
        else:
            return super(MultilinePopup, self).get_absolute_start_pos()

    def get_absolute_stop_pos(self) -> Tuple[int, int]:
        # override for custom sizes, so we can calculate the stop position based on our custom width and height instead
        # of calculating width and height based on start and stop
        if self.__custom_size:
            start_x, start_y = self.get_absolute_start_pos()
            self._stop_x, self._stop_y = start_x + self._width, start_y + self._height
            return self._stop_x, self._stop_y
        else:
            return super(MultilinePopup, self).get_absolute_stop_pos()

    def freeze(self):
        """
        The popup will be normalized to have the default size and position. Additionally, questions are transformed into
        statements with the current answer being the final/saved answer.
        """
        if self._is_question:
            # without a confirmation_callback the popup is not regarded as questions and hence the answer cannot change
            self.__confirmation_callback = None

        # reset position and size
        self.__custom_pos = PopupConfig.default_position()
        self._height, self._width = MultilinePopup.__BASE_DIMENSIONS
        # recalculate start and stop to be correct after the reset above
        self._start_x, self._start_y = self.get_absolute_start_pos()
        self._stop_x, self._stop_y = self.get_absolute_stop_pos()



import enum
from abc import ABC, abstractmethod
from typing import List, Callable, Optional, Tuple, Any

import qrogue.game.world.tiles as tiles
from qrogue.game.logic import Message
from qrogue.game.logic.actors import Controllable, Robot
from qrogue.game.world.map.rooms import Room, Area, Placeholder, SpawnRoom, MetaRoom, AreaType
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.util import Logger, MapConfig


class MapType(enum.Enum):
    World = 0
    Level = 1
    Expedition = 2


class MapMetaData:
    def __init__(self, name: Optional[str], description: Optional[Message], has_teleporter: bool,
                 show_description: Callable[[], None], show_individual_qubits: bool = False,
                 end_message: Optional[Message] = None):
        self.__name = name
        self.__description = description
        self.__has_teleporter = has_teleporter
        self.__show_description = show_description
        self.__show_individual_qubits = show_individual_qubits
        self.__end_message = end_message

    @property
    def name(self) -> Optional[str]:
        return self.__name

    @property
    def description(self) -> Optional[Message]:
        return self.__description

    @property
    def has_teleporter(self) -> bool:
        return self.__has_teleporter

    @property
    def show_description(self) -> Callable:
        return self.__show_description

    @property
    def show_individual_qubits(self) -> bool:
        return self.__show_individual_qubits

    @property
    def end_message(self) -> Optional[Message]:
        return self.__end_message


class BaseMap(ABC):
    @staticmethod
    def _calculate_pos(pos_of_room: Coordinate, pos_in_room: Coordinate) -> Coordinate:
        """
        Calculates and returns a Coordinate on the Map corresponding to the Coordinate of a Room on the Map and
        a Coordinate in the Room.

        :param pos_of_room: Coordinate of the Room on the Map
        :param pos_in_room: Coordinate in the Room
        :return: Coordinate on the Map
        """
        x = pos_of_room.x * (Area.UNIT_WIDTH + 1) + pos_in_room.x
        y = pos_of_room.y * (Area.UNIT_HEIGHT + 1) + pos_in_room.y
        return Coordinate(x, y)

    def __init__(self, meta_data: MapMetaData, internal_name: str, seed: int, rooms: List[List[Optional[Room]]]):
        self.__meta_data = meta_data
        self.__internal_name = internal_name
        self.__seed = seed
        self.__rooms = rooms
        self.__dimensions = Coordinate(len(rooms[0]), len(rooms))

    @property
    def _meta_data(self) -> MapMetaData:
        return self.__meta_data

    @property
    def _rooms(self) -> List[List[Optional[Room]]]:
        return self.__rooms

    @property
    def name(self) -> str:
        """
        Name of the Map that is presented to the player.
        Can be the same as its internal name.
        :return: name of the Map that is shown in-game
        """
        if self.__meta_data.name is None:  # show the internal name if no display name was specified
            return self.__internal_name
        return self.__meta_data.name

    @property
    def internal_name(self) -> str:
        """
        Name of the Map that is used to identify it in the game's logic (e.g. for events).
        :return: name of the Map that is used internally
        """
        return self.__internal_name

    @property
    def seed(self) -> int:
        return self.__seed

    @property
    def width(self) -> int:
        return self.__dimensions.x

    @property
    def height(self) -> int:
        return self.__dimensions.y

    @property
    def full_width(self) -> int:
        """

        :return:
        """
        return self.width * (Area.UNIT_WIDTH + 1) - 1

    @property
    def full_height(self) -> int:
        return self.height * (Area.UNIT_HEIGHT + 1) - 1

    @abstractmethod
    def get_type(self) -> MapType:
        pass

    def _get_area(self, x: int, y: int) -> Tuple[Optional[Area], tiles.Tile]:
        """
        Calculates and returns the Room and in-room Coordinates of the given Map position (globally, not room position!).
        :param x: x position on the Map
        :param y: y position on the Map
        :return: Room or Hallway and their Tile at the given position
        """
        in_hallway = None
        width = Area.UNIT_WIDTH + 1
        height = Area.UNIT_HEIGHT + 1
        x_mod = x % width
        y_mod = y % height
        # position is in Hallway
        if x_mod == Area.UNIT_WIDTH:
            if y_mod == Area.UNIT_HEIGHT:
                # there are a few points on the map that are surrounded by Hallways and don't belong to any Room
                Logger.instance().error(f"Error! You should not be able to move outside of Hallways: {x}|{y}",
                                        show=False, from_pycui=False)
                return None, tiles.Invalid()
            x -= 1
            x_mod -= 1
            in_hallway = Direction.East
        elif y_mod == Area.UNIT_HEIGHT:
            y -= 1
            y_mod -= 1
            in_hallway = Direction.South

        room_x = int(x / width)
        room_y = int(y / height)
        room = self.__rooms[room_y][room_x]
        if room is None:
            Logger.instance().error(f"Error! Invalid position: {x}|{y}", show=False, from_pycui=False)
            return None, tiles.Invalid()

        if in_hallway:
            hallway = room.get_hallway(in_hallway)
            if hallway is None:
                return None, tiles.Invalid()
            if hallway.is_horizontal():
                return hallway, hallway.at(x_mod, 0)
            else:
                return hallway, hallway.at(0, y_mod)
        else:
            return room, room.at(x_mod, y_mod)

    def room_at(self, x: int, y: int) -> Optional[Room]:
        """
        Returns the Room at the given position or None if either x or y are out of bounds.

        :param x: x-Coordinate of the room we want to get
        :param y: y-Coordinate of the room we want to get
        :return: the Room at the given position or None if the position is invalid
        """
        if 0 <= x < self.width and 0 <= y < self.height:
            return self.__rooms[y][x]
        return None

    def row_strings(self) -> List[str]:
        rows: List[str] = []
        offset = 0
        # iterate through every row of Rooms
        for y in range(self.height):
            last_row = y == self.height - 1  # there are no more Hallways after the last row of Rooms
            areas: List[Area] = []
            south_hallways: List[str] = []

            for x in range(self.width):
                last_col = x == self.width - 1  # there are no more Hallways after the last Room in a row
                room = self.room_at(x, y)
                if room is None:
                    areas.append(Placeholder.pseudo_room())
                    if not last_col:
                        areas.append(Placeholder.vertical())
                    if not last_row:
                        south_hallways.append(Placeholder.horizontal().get_row_str(0))

                else:
                    areas.append(room)
                    if not last_col:
                        hallway = room.get_hallway(Direction.East, throw_error=False)
                        if hallway is None:
                            areas.append(Placeholder.vertical())
                        else:
                            areas.append(hallway)
                    if not last_row:
                        hallway = room.get_hallway(Direction.South, throw_error=False)
                        if hallway is None:
                            south_hallways.append(Placeholder.horizontal().get_row_str(0))
                        else:
                            south_hallways.append(hallway.get_row_str(0))

            rows += ([""] * Area.UNIT_HEIGHT)  # initialize the new "block" of rows with empty strings
            for area in areas:
                for ry in range(Area.UNIT_HEIGHT):
                    rows[offset + ry] += area.get_row_str(ry)
            rows.append(Area.void().get_img().join(south_hallways))
            offset += Area.UNIT_HEIGHT + 1
        return rows

    def __str__(self):
        return "\n".join(self.row_strings())


class Map(BaseMap, ABC):
    def __init__(self, meta_data: MapMetaData, internal_name: str, seed: int, rooms: List[List[Optional[Room]]],
                 controllable: Controllable, spawn_room: Coordinate, check_achievement: Callable[[str], bool],
                 trigger_event: Callable[[str], None]):
        super().__init__(meta_data, internal_name, seed, rooms)
        self.__check_achievement = check_achievement
        self.__trigger_event = trigger_event

        self.__controllable_tile = tiles.ControllableTile(controllable)
        self.__controllable_pos = BaseMap._calculate_pos(spawn_room, Coordinate(Area.MID_X, Area.MID_Y))
        self.__move_direction = Direction.Center  # initially we didn't move in any direction

        self.__cur_area = self.room_at(spawn_room.x, spawn_room.y)
        if self.__cur_area is None:
            Logger.instance().error(f"Illegal spawn room @{spawn_room} for map {meta_data.name}", show=False,
                                    from_pycui=False)
        else:
            self.__cur_area.enter(Direction.Center)
            self.__cur_area.make_visible()

            if isinstance(self.__cur_area, SpawnRoom):
                self.__cur_area.set_is_done_callback(self._is_done)
            elif not isinstance(self.__cur_area, MetaRoom) and self.__cur_area.type is not AreaType.SpawnRoom:
                Logger.instance().error(f"{meta_data.name} starts in area that is not a SpawnRoom! cur_area = "
                                        f"{self.__cur_area}", show=False, from_pycui=False)

    @property
    def robot(self) -> Robot:
        if isinstance(self.__controllable_tile.controllable, Robot):
            return self.__controllable_tile.controllable
        raise TypeError(f"{self.__controllable_tile.controllable} is not a robot!")

    @property
    def robot_pos(self) -> Coordinate:
        return self.__controllable_pos

    @property
    def robot_img(self) -> Any:
        return self.__controllable_tile.get_img()

    @property
    def show_individual_qubits(self) -> bool:
        return self._meta_data.show_individual_qubits

    def _is_done(self) -> bool:
        return self.__check_achievement(MapConfig.done_event_id())

    def start(self):
        self._meta_data.show_description()

    def end(self, show_message: Callable[[Optional[Message]], None]):
        show_message(self._meta_data.end_message)

    def move(self, direction: Direction) -> bool:
        """
        Tries to move the robot into the given Direction.
        :param direction: in which direction the robot should move
        :return: True if the robot was able to move, False otherwise
        """
        if self.robot.game_over_check():
            # although moving does not cost energy, this can still be True if we didn't get back energy after a puzzle
            # todo fix somehow since it is awkward to immediately lose after solving a puzzle correctly
            return False

        new_pos = self.__controllable_pos + direction
        if new_pos.y < 0 or self.full_height <= new_pos.y or \
                new_pos.x < 0 or self.full_width <= new_pos.x:
            return False

        area, tile = self._get_area(new_pos.x, new_pos.y)
        if tile.is_walkable(direction, self.robot):
            self.robot.on_move()

            if area != self.__cur_area:
                self.__cur_area.leave(direction)
                self.__cur_area = area
                self.__cur_area.enter(direction)

            if isinstance(tile, tiles.WalkTriggerTile):
                tile.trigger(direction, self.robot, self.__trigger_event)

            self.__controllable_pos = new_pos
            self.__move_direction = direction
            return True
        else:
            return False

    def undo_last_move(self) -> bool:
        """
        Tries to undo the last movement (i.e., move in the opposite direction). Fails at the beginning of a Map when
        there was no movement yet or if something blocks the Controllable to move back (e.g., directional doors).

        :return: True if we were able to undo the last move, False otherwise
        """
        if self.__move_direction is Direction.Center: return False

        return self.move(self.__move_direction.opposite())

    def tunnel(self, pos_of_room: Coordinate, pos_in_room: Optional[Coordinate]) -> bool:
        if pos_in_room is None:
            _x, _y = MapConfig.room_mid()
            pos_in_room = Coordinate(_x, _y)

        room = self.room_at(pos_of_room.x, pos_of_room.y)
        if room is None:
            return False

        target_pos = BaseMap._calculate_pos(pos_of_room, pos_in_room)
        direction = Direction.from_coordinates(self.__controllable_pos, target_pos)
        destination_tile = room.at(pos_in_room.x, pos_in_room.y, force=True)
        if destination_tile.is_walkable(direction, self.robot):
            self.__controllable_pos = target_pos
            return True
        return False

    def __str__(self):
        return "\n".join(self.row_strings())



import math
from typing import List, Optional, Tuple, Dict, Set

from qrogue.game.logic.base import StateVector
from qrogue.game.logic.collectibles import Collectible, Instruction
from qrogue.game.logic.collectibles.instruction import RZGate, RYGate
from qrogue.util import MyRandom, RandomManager, DifficultyType, StvDifficulty, Config, Logger


class PuzzleGenerator:
    __MAX_TRIES = 5

    @staticmethod
    def __prepare_gate_at(rm: MyRandom, circuit_space: int, gate: Instruction, qubit_count: List[int],
                          qubits: List[int], qubit: int) -> bool:
        """
        :returns: True if gate was successfully prepared, False otherwise
        """
        gate_qubits = qubits.copy()
        assert qubit in gate_qubits, f"Invalid arguments: qubit={qubit} not in qubits={qubits}!"
        gate_qubits.remove(qubit)

        qubit_count[qubit] += 1
        if qubit_count[qubit] >= circuit_space:
            qubits.remove(qubit)

        if gate.use_qubit(qubit):
            if len(qubits) <= 0:
                Logger.instance().error(f"Not enough qubits provided to prepare gate {gate.name()}!", from_pycui=False)
                return False
            # don't remove here because removal happens in the next recursion step (because we need to pass a qubit
            #  that is still in qubits for the very first call)
            next_qubit = rm.get_element(gate_qubits, remove=False, msg="BossFactory_selectQubit")
            return PuzzleGenerator.__prepare_gate_at(rm, circuit_space, gate, qubit_count, gate_qubits, next_qubit)
        return True

    @staticmethod
    def __get_angle(rm: MyRandom, randomization_degree: int) -> float:
        if randomization_degree == 0:
            angle = rm.get(msg="PuzzleGenerator.produce()_unrestrictedAngle")
        else:  # todo: what about randomization_degree==1?
            # start with 1 because rotating by 0° does nothing (since range-end is exclusive, 360° can't happen)
            angle = rm.get_element([(2 * math.pi) * i / randomization_degree
                                    for i in range(1, randomization_degree)],
                                   msg="PuzzleGenerator.produce()_restrictedAngle")
        return angle

    @staticmethod
    def prepare_single_layer_gates(rm: MyRandom, num_of_qubits: int, circuit_space: int, difficulty: StvDifficulty,
                                   available_gates: List[Instruction]) -> List[Instruction]:
        """
        Prepares at most one single-qubit gate per qubit.

        :param rm: determines randomness
        :param num_of_qubits: how many qubits can be used to place gates on
        :param circuit_space: the maximum number of gates that can be placed in the circuit
        :param difficulty: determines absolute value of QubitExuberance, and hence, how many gates we try to prepare
                            relative to num_of_qubits
        :param available_gates: the gates available for preparation
        :returns: list of prepared (i.e., ready to compute a StateVector) gates
        """
        gate_list = []
        # 1) extract difficulty value
        num_of_single_layer_qubits = difficulty.get_absolute_value(DifficultyType.QubitExuberance, num_of_qubits,
                                                                   circuit_space)
        # 2) decide on which qubits we apply a single gate
        qubit_count = [0] * num_of_qubits
        qubits = list(range(num_of_qubits))
        single_layer_qubits = [rm.get_element(qubits, remove=True) for _ in range(num_of_single_layer_qubits)]

        # 3) filter for single qubit gates
        filtered_gates = [gate for gate in available_gates if gate.num_of_qubits == 1]

        # 4) apply single qubit gates to the selected qubits
        for qubit in single_layer_qubits:
            if len(filtered_gates) <= 0: break
            gate = rm.get_element(filtered_gates, remove=True)
            if PuzzleGenerator.__prepare_gate_at(rm, circuit_space, gate, qubit_count, [qubit], qubit):
                gate_list.append(gate)

        return gate_list

    @staticmethod
    def prepare_rotation_gates(rm: MyRandom, num_of_qubits: int, circuit_space: int, difficulty: StvDifficulty) \
            -> List[Instruction]:
        rotation_gates: List[Instruction] = []

        # 1) retrieve variables
        ############################
        diff_values = difficulty.get_absolute_dict(num_of_qubits, circuit_space)
        num_of_rotated_qubits = diff_values[DifficultyType.QubitExuberance]
        rotation_degree = diff_values[DifficultyType.RotationExuberance]
        randomization_degree = diff_values[DifficultyType.RandomizationDegree]
        ############################

        # 2) decide which qubits should receive rotations
        ############################
        qubit_count = [0] * num_of_qubits
        qubits = list(range(num_of_qubits))

        # get random qubits to fill the rotated_qubits list
        rotated_qubits = [rm.get_element(qubits, remove=True) for _ in range(num_of_rotated_qubits)]

        if rotation_degree > 0:
            rotated_qubits2 = rotated_qubits.copy()
            # randomly remove qubits until only #rotation_degree qubits are left in rotated_qubits2
            for _ in range(len(rotated_qubits) - rotation_degree):
                rm.get_element(rotated_qubits2, remove=True)
        else:
            # no qubit receives a second rotation
            rotated_qubits2 = []
        ############################

        # 3) apply rotation gates to the 0-basis state based on previously selected qubits
        ############################
        for qubit in rotated_qubits:
            # find a random angle for the rotation
            angle = PuzzleGenerator.__get_angle(rm, randomization_degree)

            # prepare and append either an RY- or RZ-Gate
            rotate_y = rm.get_bool(msg="PuzzleGenerator.produce()_chooseRGate")
            rot_gate1 = RYGate(angle) if rotate_y else RZGate(angle)
            if PuzzleGenerator.__prepare_gate_at(rm, circuit_space, rot_gate1, qubit_count, [qubit], qubit):
                rotation_gates.append(rot_gate1)

            # append the other rotational gate if the qubit is also in rotated_qubits2 (i.e., it receives two rotations)
            if qubit in rotated_qubits2:
                # now use RZGate if rotate_y is True (RYGate was used before), and vice versa
                rot_gate2 = RZGate(angle) if rotate_y else RYGate(angle)
                if PuzzleGenerator.__prepare_gate_at(rm, circuit_space, rot_gate2, qubit_count, [qubit], qubit):
                    rotation_gates.append(rot_gate2)
        ############################

        return rotation_gates

    @staticmethod
    def prepare_from_gates(rm: MyRandom, num_of_qubits: int, circuit_space: int, difficulty: StvDifficulty,
                           available_gates: List[Instruction], include_gates: Optional[List[Instruction]],
                           force_num_of_gates: bool = False, inverse: bool = False) -> List[Instruction]:
        available_gates = available_gates.copy()  # copy it because we might remove elements during selection

        # 1) prepare variables
        ##################################
        if include_gates is None: include_gates = []
        num_of_gates = difficulty.get_absolute_value(DifficultyType.CircuitExuberance, num_of_qubits, circuit_space)

        if len(include_gates) > num_of_gates:
            Logger.instance().warn(f"Cannot use all included gates: only {num_of_gates} out of {len(include_gates)} "
                                   f"will be used.", from_pycui=False)
        ##################################

        gate_list = []
        prepare_rm = RandomManager.create_new(rm.get_seed("PrepareRM"))
        for _ in range(PuzzleGenerator.__MAX_TRIES):    # try multiple times if we fail to generate a gate_list
            l_available_gates = available_gates.copy()  # copy available_gates since we remove elements later

            # 2) select the gates to choose for target
            ##################################
            if len(include_gates) + len(l_available_gates) <= num_of_gates:
                # we don't have enough gates to be picky, so we choose all we can
                selected_gates = include_gates + l_available_gates
            else:
                selected_gates = include_gates.copy()  # include_gates need to be used
                while len(selected_gates) < num_of_gates:
                    # add random ones from available_gates to the selection
                    selected_gates.append(rm.get_element(l_available_gates, remove=True,
                                                         msg="PuzzleGenerator.prepare_target()@gates_for_target"))
            rm.shuffle_list(selected_gates)
            ##################################

            # 3) prepare gate_list
            ##################################
            if force_num_of_gates:
                gate_list = PuzzleGenerator.__prepare_nearsighted_circuit(prepare_rm, num_of_qubits, circuit_space,
                                                                          selected_gates, inverse)
                # in this case it is fine as long as we prepared at least one gate
                if len(gate_list) > 0: break
            else:
                gate_list = PuzzleGenerator.__prepare_farsighted_circuit(prepare_rm, num_of_qubits, circuit_space,
                                                                         selected_gates, inverse)
                # in this case we need to prepare at least half the expected number of gates
                if len(gate_list) >= len(selected_gates) / 2: break
            ##################################

            # 4) retry if needed
            Config.check_reachability("Retry gate_list generation.")
            prepare_rm = RandomManager.create_new(rm.get_seed("PrepareRM-while"))
        return gate_list

    @staticmethod
    def __prepare_nearsighted_circuit(rm: MyRandom, num_of_qubits: int, circuit_space: int,
                                      selected_gates: List[Instruction], inverse: bool = False) -> List[Instruction]:
        """
        Prepares a list of gates (representing a circuit) such that any two consecutive gates on a qubit don't cancel
        each other out. It is still possible that, e.g., the gate placed third cancels the combination of the first and
        second gate.
        """
        qubit_count = [0] * num_of_qubits
        qubits = list(range(num_of_qubits))

        # store which stvs we generated on the way to make sure we don't undo a previous gate (leading to virtually
        #  fewer gates used than tasked with) - starting with no gates applied, we're at the 0-state
        prev_stvs: Set[StateVector] = {StateVector.create_zero_state_vector(num_of_qubits)}
        circuits: List[List[Instruction]] = [[]]

        i, start_index = 0, 0
        while True:
            added_gate = False
            cur_qubits = qubits.copy()
            rm.shuffle_list(cur_qubits)
            # try to add the currently selected gate on different qubits until the resulting stv is completely new
            for qu in cur_qubits:
                cur_gate = selected_gates[i].copy()
                if not PuzzleGenerator.__prepare_gate_at(rm, circuit_space, cur_gate, qubit_count, qubits, qu):
                    continue    # failed to prepare the gate, so let's try with a different qubit

                real_circuit = circuits[-1]
                new_stv = Instruction.compute_stv(real_circuit.copy() + [cur_gate], num_of_qubits, inverse)

                # check whether new_stv is indeed new or a repeat of a previous one (meaning the gate undid an operation)
                is_repeated = False
                for prev_stv in prev_stvs:
                    if prev_stv.is_equal_to(new_stv, ignore_god_mode=True):
                        is_repeated = True  # we found a duplicate
                        break

                if not is_repeated:
                    # gates_for_target and cur_gate result in a new stv -> save cur_gate to gates_for_target
                    # gates_for_target.append(cur_gate)
                    # prev_stvs.append(new_stv)  # also store the resulting stv for comparison with new ones
                    new_circuits = []
                    for circuit in circuits:
                        new_circuit = circuit.copy() + [cur_gate]
                        new_circuits.append(new_circuit)
                        prev_stvs.add(Instruction.compute_stv(new_circuit, num_of_qubits, inverse))
                    circuits += new_circuits

                    added_gate = True
                    break

            # ret_val = PuzzleGenerator.__next_gate(i, start_index, added_gate, selected_gates)
            # if ret_val is None: break
            # i, start_index, selected_gates = ret_val
            if added_gate:
                # remove cur_gate (at index i) from selectable gates
                if len(selected_gates) == 1:
                    break  # we just placed the last gate
                elif i + 1 >= len(selected_gates):
                    selected_gates = selected_gates[:i]
                else:
                    selected_gates = selected_gates[:i] + selected_gates[i + 1:]
                # don't increment i since it already points to a new gate
                if i >= len(selected_gates): i = 0  # change i by starting over if it would be out of bounds
                start_index = i  # store where the search for the next gate started, so we can stop before repeating
            else:
                i += 1  # try the next gate
                if i >= len(selected_gates): i = 0
                if i == start_index:
                    Config.check_reachability("target_stv generation unsuccessful")
                    break  # we couldn't add all selected gates

        return circuits[-1]

    @staticmethod
    def __prepare_farsighted_circuit(rm: MyRandom, num_of_qubits: int, circuit_space: int,
                                     selected_gates: List[Instruction], inverse: bool = False) -> List[Instruction]:
        """
        Prepares a list of gates (representing a circuit) such that gate cancels out any previously placed gates. This
        way it is not ensured that all gates in selected_gates are in the circuit, but adding one of these remaining
        gates to any qubit would result in a circuit that either produces the same StateVector or a StateVector closer
        to the 0-state (i.e., undoing one of the previous gates).
        """
        # todo: a bad gate selection can lead to being unable to place enough gates - fix with additional parameters and
        #  selecting gates on the fly? (but we have to make sure that included_gates are used!)
        # todo: maybe we can
        #  change included_gates to a single gate (ComposedGates would still allow to use multiple)
        qubit_count = [0] * num_of_qubits
        qubits = list(range(num_of_qubits))

        last_gate_dict: Dict[int, Optional[Instruction]] = {}
        for cur_qubit in qubits: last_gate_dict[cur_qubit] = None

        circuit = []
        i, start_index = 0, 0
        while True:
            added_gate = False
            cur_qubits = qubits.copy()
            rm.shuffle_list(cur_qubits)
            # try to add the currently selected gate on different qubits until the resulting stv is completely new
            for cur_qubit in cur_qubits:
                cur_gate = selected_gates[i].copy()
                if not PuzzleGenerator.__prepare_gate_at(rm, circuit_space, cur_gate, qubit_count, qubits, cur_qubit):
                    continue    # failed to prepare the gate, so let's try with a different qubit

                # if stv of circuit together with cur_gate would result in a 0-state, we continue with the next qubit
                if Instruction.compute_stv(circuit + [cur_gate], num_of_qubits, inverse).is_zero: continue

                # check if the last gate on cur_qubit is cancelled by cur_gate
                if last_gate_dict[cur_qubit] is not None and \
                        Instruction.compute_stv([last_gate_dict[cur_qubit], cur_gate], num_of_qubits, inverse).is_zero:
                    continue

                # circuit and cur_gate result in a new stv -> save cur_gate to circuit
                circuit.append(cur_gate)
                # update last_gate_dict based on newly appended cur_gate
                for qu in cur_gate.qargs_iter(): last_gate_dict[qu] = cur_gate
                # set added_gate flag, so we now that we can remove cur_gate from selected_gates
                added_gate = True
                break

            # ret_val = PuzzleGenerator.__next_gate(i, start_index, added_gate, selected_gates)
            # if ret_val is None: break
            # i, start_index, selected_gates = ret_val
            if added_gate:
                # remove cur_gate (at index i) from selectable gates
                if len(selected_gates) == 1:
                    break  # we just placed the last gate
                elif i + 1 >= len(selected_gates):
                    selected_gates = selected_gates[:i]
                else:
                    selected_gates = selected_gates[:i] + selected_gates[i + 1:]
                # don't increment i since it already points to a new gate
                if i >= len(selected_gates): i = 0  # change i by starting over if it would be out of bounds
                start_index = i  # store where the search for the next gate started, so we can stop before repeating
            else:
                i += 1  # try the next gate
                if i >= len(selected_gates): i = 0
                if i == start_index:
                    Config.check_reachability("target_stv generation unsuccessful")
                    break  # we couldn't add all selected gates

        return circuit

    @staticmethod
    def __next_gate(index: int, start_index: int, added_gate: bool, gate_list: List[Instruction]) \
            -> Optional[Tuple[int, int, List[Instruction]]]:  # todo: fix and use
        if added_gate:
            # remove cur_gate (at index i) from selectable gates
            if len(gate_list) == 1:
                return None  # we just placed the last gate
            elif index + 1 >= len(gate_list):
                selected_gates = gate_list[:index]
            else:
                selected_gates = gate_list[:index] + gate_list[index + 1:]
            # don't increment i since it already points to a new gate
            if index >= len(selected_gates): i = 0  # change i by starting over if it would be out of bounds
            start_index = index  # store where the search for the next gate started, so we can stop before repeating
        else:
            index += 1  # try the next gate
            if index >= len(gate_list): index = 0
            if index == start_index:
                Config.check_reachability("target_stv generation unsuccessful")
                return None  # we couldn't add all selected gates

        return index, start_index, gate_list

    @staticmethod
    def generate_puzzle(rm: MyRandom, num_of_qubits: int, circuit_space: int, difficulty: StvDifficulty,
                        available_gates: List[Instruction], include_gates: Optional[List[Instruction]],
                        reward_pool: List[Collectible]) -> Tuple[StateVector, StateVector, Optional[Collectible]]:
        irm = RandomManager.create_new(rm.get_seed())  # input rm
        trm = RandomManager.create_new(rm.get_seed())  # target rm
        rrm = RandomManager.create_new(rm.get_seed())  # reward rm

        input_gates = PuzzleGenerator.prepare_rotation_gates(irm, num_of_qubits, circuit_space, difficulty)
        input_stv = Instruction.compute_stv(input_gates, num_of_qubits)

        target_gates = PuzzleGenerator.prepare_from_gates(trm, num_of_qubits, circuit_space, difficulty,
                                                          available_gates, include_gates)
        target_stv = Instruction.compute_stv(input_gates + target_gates, num_of_qubits)

        reward = rrm.get_element(reward_pool, msg="generate_puzzle()-reward")

        return input_stv, target_stv, reward



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

from qrogue.game.logic.collectibles import CollectibleType
from qrogue.game.world import tiles
from qrogue.game.world.map import rooms, LessonMap
from qrogue.game.world.navigation import Coordinate
from qrogue.util import Logger
from qrogue.util.util_functions import my_str, enum_from_string


class WFCLearnMatrix(ABC):
    @property
    @abstractmethod
    def width(self) -> int:
        pass

    @property
    @abstractmethod
    def height(self) -> int:
        pass

    @property
    @abstractmethod
    def type(self) -> Any:
        pass

    @abstractmethod
    def at(self, x: int, y: int) -> Any:
        """

        :param x:
        :param y:
        :return: a value of the feature we want to learn about
        """
        pass


class LearnableMap(WFCLearnMatrix):
    def __init__(self, level_map: LessonMap):
        self.__level_map = level_map

    @property
    def width(self) -> int:
        return self.__level_map.width

    @property
    def height(self) -> int:
        return self.__level_map.height

    @property
    def type(self) -> Any:
        return rooms.AreaType

    def at(self, x: int, y: int) -> rooms.AreaType:
        room = self.__level_map.room_at(x, y)
        if room is None:
            return rooms.AreaType.Invalid
        return room.type


class LearnableRoom(WFCLearnMatrix):
    class TileData:
        __DATA_PREFIX = " ("
        __DATA_SUFFIX = ")"

        @staticmethod
        def from_string(tile_data: str) -> Optional["LearnableRoom.TileData"]:
            # retrieve data if present
            prefix = LearnableRoom.TileData.__DATA_PREFIX
            if prefix in tile_data:
                data_start = tile_data.index(prefix) + len(prefix)
                data_end = tile_data.index(LearnableRoom.TileData.__DATA_SUFFIX, data_start)
                data = tile_data[data_start:data_end]
                tile_data = tile_data[:data_start-len(prefix)]  # don't include prefix
            else:
                data = None

            # parse TileCode
            tile_code = enum_from_string(tiles.TileCode, tile_data)
            if tile_code is tiles.TileCode.Enemy:
                if data is None:
                    Logger.instance().warn(f"Tried to create TileData for Enemy without specified data: {tile_data}",
                                           from_pycui=False)
                    return None
                data = int(data)    # enemies use their id as data

            elif tile_code is tiles.TileCode.Collectible:
                if data is None:
                    Logger.instance().warn(f"Tried to create TileData for unspecified Collectible without data: "
                                           f"{tile_data}", from_pycui=False)
                    return None
                data = enum_from_string(CollectibleType, data)

            return LearnableRoom.TileData(tile_code, data)

        def __init__(self, code: tiles.TileCode, data: Any):
            self.__code = code
            self.__data = data

            if data is not None and not isinstance(data, int) and not isinstance(data, CollectibleType):
                from qrogue.util import Config
                Config.check_reachability(f"unexpected type for data: {data}")

        @property
        def code(self) -> tiles.TileCode:
            return self.__code

        @property
        def data(self) -> Any:
            return self.__data

        def __hash__(self):
            if self.__data is None:
                return hash(self.__code)
            else:
                return hash(self.__code) * 17 + hash(self.__data)

        def __eq__(self, other):
            if isinstance(other, tiles.TileCode):
                return other == self.__code
            elif isinstance(other, LearnableRoom.TileData):
                return other.__code == self.__code and other.__data == self.__data
            else:
                return False

        def __str__(self):
            return f"{self.__code.name}" + \
                ("" if self.__data is None else
                 f"{LearnableRoom.TileData.__DATA_PREFIX}{my_str(self.__data)}{LearnableRoom.TileData.__DATA_SUFFIX}")

    def __init__(self, room: rooms.Room, convert_keys: bool = True):
        """
        :param room: the room to learn from
        :param convert_keys: whether keys should be learned as universal collectibles or straight as keys (because
            expeditions usually only have one boss key)
        """
        self.__room = room
        self.__convert_keys = convert_keys

    @property
    def width(self) -> int:
        return rooms.Room.INNER_WIDTH

    @property
    def height(self) -> int:
        return rooms.Room.INNER_HEIGHT

    @property
    def type(self) -> Any:
        return tiles.TileCode

    def at(self, x: int, y: int) -> TileData:
        tile = self.__room.at(x, y, force=True, inside_only=True)
        if self.__convert_keys and tile.data is CollectibleType.Key:
            return LearnableRoom.TileData(tile.code, CollectibleType.Pickup)
        return LearnableRoom.TileData(tile.code, tile.data)


class LearnableRoomMatrix(WFCLearnMatrix):
    def __init__(self, room_dict: Dict[Coordinate, LearnableRoom.TileData]):
        self.__matrix = room_dict
        self.__width = 0
        self.__height = 0

        for coor in room_dict.keys():
            if coor.x > self.__width:
                self.__width = coor.x
            if coor.y > self.__height:
                self.__height = coor.y

        # increment width and height since coordinates start at 0
        self.__width += 1
        self.__height += 1

    @property
    def width(self) -> int:
        return self.__width

    @property
    def height(self) -> int:
        return self.__height

    @property
    def type(self) -> Any:
        return tiles.TileCode

    def at(self, x: int, y: int) -> LearnableRoom.TileData:
        coor = Coordinate(x, y)
        if coor in self.__matrix:
            return self.__matrix[Coordinate(x, y)]
        else:
            return LearnableRoom.TileData(tiles.TileCode.Invalid, None)



from typing import List, Callable, Optional, Any, Tuple

from qrogue.game.logic.actors import Controllable
from qrogue.game.logic.collectibles import Instruction
from qrogue.game.world.navigation import Coordinate
from qrogue.util import MapConfig, StvDifficulty, RandomManager, GateType, DifficultyType
from .map import Map, MapType, Room, MapMetaData


class ExpeditionMap(Map):
    # chosen characters:
    #  - Q: because quantum (disregarded for the other criteria)
    #  - (almost) the same number of consonants (13) and vowels (13)
    #  - (almost) the same number of lower (12) and upper (14) case letters
    #  - characters that look cool/spacy appear more often while maintaining the other criteria
    __NAME_CHARS = "XXxZzBFHTYcvwAaEeIIiUuuOooQQQ"

    @staticmethod
    def __get_character(value: int, offset: Optional[int] = None) -> str:
        if offset is not None: value += offset
        value %= len(ExpeditionMap.__NAME_CHARS)
        return ExpeditionMap.__NAME_CHARS[value]

    @staticmethod
    def to_display_name(difficulty: StvDifficulty, seed: int) -> str:
        rm = RandomManager.create_new(seed)
        display_name = "Expedition "
        index = 0
        for i, diff_level in enumerate(difficulty.to_code()):
            # offset makes it so the next character is definitely different
            offset = rm.get_int(min_=1, max_=len(ExpeditionMap.__NAME_CHARS) - StvDifficulty.num_of_difficulty_levels())
            character = ExpeditionMap.__get_character(index + int(diff_level), offset)
            index += StvDifficulty.num_of_difficulty_levels() + offset
            # the first letter is always upper case
            if i == 0: display_name += character.upper()
            else: display_name += character
        return display_name + f"{difficulty.level}*{seed}"

    @staticmethod
    def validate_gates_for_difficulty(difficulty: StvDifficulty, gates: List[Instruction]) -> Tuple[int, Any]:
        """
        Return codes:
            - 0 = valid
            - 1 = not enough gates, returns (needed number of gates, actual number of gates)
            - 2 = not enough unique gates, returns (needed number of unique gates, actual unique gates)
            - 3 = gates are not difficult enough, returns (needed difficulty sum, actual difficulty sum)

        :param difficulty: the StvDifficulty to validate for
        :param gates: a list of Instructions we want to validate
        :return: (0, None) if gates are valid for the given difficulty, other values depending on what is invalid

        """
        needed_num_of_gates = difficulty.get_absolute_value(DifficultyType.MinAvailableGates, 0, 0)
        num_of_gates = len(gates)
        if num_of_gates < needed_num_of_gates:
            return 1, (needed_num_of_gates, num_of_gates)

        needed_unique_gates = difficulty.get_absolute_value(DifficultyType.MinGateVariety, 0, 0)
        unique_gates = set(gates)
        if len(unique_gates) < needed_unique_gates:
            return 2, (needed_unique_gates, unique_gates)

        needed_diff_sum = difficulty.get_absolute_value(DifficultyType.MinGateDifficulty, 0, 0)
        difficulty_sum = sum([gate.difficulty for gate in gates])
        if difficulty_sum < needed_diff_sum:
            return 3, (needed_diff_sum, difficulty_sum)

        return 0, None

    def __init__(self, seed: int, difficulty: StvDifficulty, main_gate: GateType, rooms: List[List[Optional[Room]]],
                 controllable: Controllable, spawn_room: Coordinate, check_achievement: Callable[[str], bool],
                 trigger_event: Callable[[str], None]):
        # expeditions don't have a description to show, so we pass an empty lambda
        meta_data = MapMetaData(ExpeditionMap.to_display_name(difficulty, seed), None, True, lambda: None)
        self.__difficulty = difficulty
        self.__main_gate = main_gate
        super().__init__(meta_data, f"{MapConfig.expedition_map_prefix()}{difficulty.to_code()}", seed, rooms,
                         controllable, spawn_room, check_achievement, trigger_event)

    @property
    def difficulty(self) -> StvDifficulty:
        return self.__difficulty

    @property
    def main_gate(self) -> GateType:
        return self.__main_gate

    def get_type(self) -> MapType:
        return MapType.Expedition




__PREFIX = "def __init__(self,"


def gen_class(class_name: str, constructor: str) -> str:
    og_constructor = constructor    # store for debugging
    start = constructor.index(__PREFIX)
    constructor = constructor[start:]   # get rid of everything before the constructor
    if not constructor.endswith(":"): constructor += ":"
    end = constructor.rfind("):")
    parameters = constructor[len(__PREFIX):end].split(",")

    class_text = f"class {class_name}:\n"
    constructor_lines = []
    property_lines = []
    for val in parameters:
        property_lines.append("@property")
        if ":" in val:
            type_start = val.index(":")
            param_name = val[:type_start].strip()
            param_type = val[type_start+1:].strip()
            property_lines.append(f"def {param_name}(self) -> {param_type}:")
        else:
            param_name = val.strip()
            property_lines.append(f"def {param_name}(self):")
        constructor_lines.append(f"self.__{param_name} = {param_name}")
        property_lines.append(f"\treturn self.__{param_name}")
        property_lines.append("")

    constructor_text = f"\t{constructor}\n" \
                       + "\n".join([f"\t\t{ctr}" for ctr in constructor_lines]) \
                       + "\n\n"

    property_text = "\t" + "\n\t".join(property_lines)

    return class_text + constructor_text + property_text


if __name__ == "__main__":
    t = gen_class("MapMetaData",
                  "def __init__(self, name: Optional[str], description: Optional[Message], has_teleporter: bool, show_description: Callable, show_individual_qubits: bool = False, end_message: Optional[Message] = None)")
    print(t)
